from PymoNNto.Exploration.UI_Base import *
from PymoNNto.Exploration.StorageManager.StorageManager import *
from PymoNNto.Exploration.Evolution.SSH_Functions import *
from PymoNNto.Exploration.Evolution.EvolutionPlots import *
from multiprocessing import Process, Queue, Pipe
import shutil
import os

class QHLine(QFrame):
    def __init__(self):
        super(QHLine, self).__init__()
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)

class QVLine(QFrame):
    def __init__(self):
        super(QVLine, self).__init__()
        self.setFrameShape(QFrame.VLine)
        self.setFrameShadow(QFrame.Sunken)

def evolution_thread_worker(name, folder, conn):
    print('local thread started')
    try:
        dir = os.getcwd()
        os.chdir(get_epc_folder(folder)+'/'+name)
        print('move to', get_epc_folder(folder)+'/'+name)

        system_str = sys.platform
        if 'darwin' in system_str:#mac os
            os.system('python3 execute.py')
        elif 'win' in system_str:#windows
            os.system('start /wait cmd /c python3 execute.py')
        else:
            os.system('gnome-terminal -x python3 execute.py')

        os.chdir(dir)

        conn.send('execution finished')
    except:
        conn.send('thread error')
    conn.send('idle')


def copy_project(evo_name, device_str, folder):
    if device_str == 'local':
        clone_project(evo_name, folder)

    if 'ssh' in device_str:
        transfer_project(evo_name, device_str)

        Data_Folder = get_data_folder()
        if Data_Folder != './Data':
            epc_folder = Data_Folder + '/'+folder
            create_folder_if_not_exist(epc_folder)
            run_folder = epc_folder + '/' + evo_name
            create_folder_if_not_exist(run_folder)


class Execution_Manager_UI_Base(UI_Base):

    def add_ui_elements(self, left_vertical_layout, right_vertical_layout):
        return

    def On_Tab_Changed(self, i):
        return

    def get_help_txt(self):
        return '''...
        '''

    def create_execution_file(self, name, folder, file):
        return False


    def save_configuration(self,ssm):#SimpleStorageManager
        return


    def valid_configuration(self):
        return True

    def add_additional_tab_elements(self, tab, name):
        return

    def get_folder(self):
        return 'Execution_Project_Clones'

    def get_title(self):
        return 'Execute and Transfer'

    def __init__(self):
        super().__init__(title=self.get_title(), create_sidebar=True)

        self.folder = self.get_folder()

        self.tabs.currentChanged.connect(self.On_Tab_Changed)


        #self.Add_element(QLabel('Task name:'), sidebar = True, stretch=1)
        #self.task_name_edit = self.Add_element(QLineEdit('task_name'), sidebar = True, stretch=1)

        #remove...
        horizontal_layout = QHBoxLayout()
        self.sidebar.get_layout().addLayout(horizontal_layout)

        left_vertical_layout = QVBoxLayout()
        horizontal_layout.addLayout(left_vertical_layout)
        left_vertical_layout.addWidget(QLabel('Task name'))
        left_vertical_layout.addWidget(QLabel('Device'))


        self.task_name_edit = QLineEdit('task_name')
        right_vertical_layout = QVBoxLayout()
        horizontal_layout.addLayout(right_vertical_layout)
        right_vertical_layout.addWidget(self.task_name_edit)

        #self.Add_element(QLabel('Device:'), sidebar = True, stretch=1)
        ssm = SimpleStorageManager(get_epc_folder() + '/')
        servers = eval(ssm.load_param('servers', default="[]"))
        self.listwidget2 = QComboBox()#self.Add_element(, sidebar=True, stretch=1)

        #self.listwidget2.SizeAdjustPolicy(3)

        self.listwidget2.addItems(['local'] + servers)#'+'
        #self.listwidget2.currentItemChanged.connect(self.on_server_select)

        server_horizontal_layout = QHBoxLayout()
        server_horizontal_layout.addWidget(self.listwidget2, stretch=10)

        add_btn = QPushButton("+")
        add_btn.clicked.connect(self.on_server_add)
        add_btn.setMaximumWidth(30)
        server_horizontal_layout.addWidget(add_btn, stretch=0)

        del_btn = QPushButton("-")
        del_btn.clicked.connect(self.on_server_delete)
        del_btn.setMaximumWidth(30)
        server_horizontal_layout.addWidget(del_btn, stretch=0)

        right_vertical_layout.addLayout(server_horizontal_layout)



        #self.Add_element(self.listwidget2, stretch=1, sidebar=True)

        self.add_ui_elements(left_vertical_layout, right_vertical_layout)

        self.add_btn = self.sidebar.add_widget(QPushButton('Add'), stretch=1)
        #self.add_btn = self.Add_element(QPushButton('Add'), sidebar=True, stretch=1)
        self.add_btn.clicked.connect(self.on_add_click)


        def help(event):
            layout = QVBoxLayout()
            pte = QPlainTextEdit()
            pte.setPlainText(self.get_help_txt())
            pte.setReadOnly(True)
            layout.addWidget(pte)

            # self.layout.addWidget(message)
            # self.layout.addWidget(self.buttonBox)

            dlg = QDialog()
            dlg.setWindowTitle("Help")
            dlg.setLayout(layout)
            dlg.resize(1000, 700)
            dlg.exec()

        help_btn = self.sidebar.add_widget(QLabel('Click for help'), stretch=1)
        help_btn.setStyleSheet('color: gray')
        help_btn.setAlignment(Qt.AlignCenter)
        help_btn.mouseReleaseEvent = help

        def on_restore_click():
            if self.archive_list.currentItem() is not None:
                file = self.archive_list.currentItem().text()
                if '.zip' in file:
                    zfile = get_epc_folder(self.folder) + '/' + file
                    zip_file = zipfile.ZipFile(zfile, mode='r')
                    zip_file.extractall(path=get_epc_folder(self.folder) + '/' + file.replace('.zip', '')  + '/')
                    zip_file.close()

                    self.add_common_tab(file.replace('.zip', ''))

                    os.remove(zfile)
                    self.update_archive()

        def on_delete_click():
            if self.archive_list.currentItem() is not None:
                file = self.archive_list.currentItem().text()
                if '.zip' in file:
                    zfile = get_epc_folder(self.folder) + '/' + file
                    os.remove(zfile)
                    self.update_archive()

        self.add_tab('...', stretch=0)
        self.archive_list = self.tab.add_widget(QLabel('Archive:'))
        self.tab.add_row(stretch=100)
        self.archive_list = self.tab.add_widget(QListWidget())
        self.tab.add_row(stretch=0)
        restore_btn = self.tab.add_widget(QPushButton('restore'))
        restore_btn.clicked.connect(on_restore_click)

        delete_btn = self.tab.add_widget(QPushButton('delete'))
        delete_btn.clicked.connect(on_delete_click)


        self.update_archive()

        for dir in os.listdir(get_epc_folder(self.folder)):
            if os.path.isdir(get_epc_folder(self.folder) + '/' + dir):
                self.add_common_tab(dir)

    def update_archive(self):
        self.archive_list.clear()
        for dir in os.listdir(get_epc_folder(self.folder)):
            if not os.path.isdir(get_epc_folder(self.folder) + '/' + dir) and '.zip' in dir:
                self.archive_list.addItem(dir)

    def check_file(self):
        file = self.slave_file_edit.text()
        root = get_root_folder()
        if root in file:
            file = file.replace(root + '/', '')
            self.slave_file_edit.setText(file)
        else:
            if os.path.isfile(root+'/'+file):
                self.slave_file_edit.setStyleSheet("background-color: rgb(255, 255, 255);")
            else:
                self.slave_file_edit.setStyleSheet("background-color: rgb(255, 150, 150);")

    def select_file(self):
        root=get_root_folder()
        dialog = QFileDialog()
        dialog.setWindowTitle('Select evolvable file')
        dialog.setNameFilter('Python Files (*.py)')
        dialog.setDirectory(root)#QtCore.QDir.currentPath()
        dialog.setFileMode(QtWidgets.QFileDialog.ExistingFile)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            file=str(dialog.selectedFiles()[0])
            if root in file:
                file=file.replace(root+'/','')
                self.slave_file_edit.setText(file)
            else:
                self.show_message('Error', 'file not in project directory')
                print('file not in project directory')
        else:
            print('no file selected')

    #def keyPressEvent(self, event):
    #    if event.key() in [Qt.Key_I]:
    #        tab = self.tabs.currentWidget()
    #        print(tab.clicked_id)
    #        smg = StorageManagerGroup(tab.name, data_folder=get_data_folder() + '/'+self.folder+'/' + tab.name + '/Data')
    #        smg.sort_by('generation')
    #        sm = smg.StorageManagerList[tab.clicked_id]
    #        file = sm.absolute_path + sm.config_file_name
    #        print(file)
    #        with open(file) as f:
    #            lines = f.readlines()

    #            layout = QVBoxLayout()
    #            pte = QPlainTextEdit()
    #            pte.setPlainText('\r\n'.join(lines))
    #            pte.setReadOnly(True)
    #            layout.addWidget(pte)
    #            dlg = QDialog()
    #            dlg.setWindowTitle("Info")
    #            dlg.setLayout(layout)
    #            dlg.resize(600, 400)
    #            dlg.exec()

    def set_text(self, ssm, edit, text):
        data = ssm.load_param(text, default='', return_string=True)
        if data is not None and data != '':
            edit.setText(data)

    def update_status(self, tab, name):

        if os.path.exists(tab.ssm.get_path()+'Data/error.txt'):
            tab.warning_label.setVisible(True)

            def show_error_msg(event):
                layout = QVBoxLayout()
                pte = QPlainTextEdit()
                pte.setPlainText(tab.ssm.load_str('Data/error'))
                pte.setReadOnly(True)
                layout.addWidget(pte)

                dlg = QDialog()
                dlg.setWindowTitle("Error")
                dlg.setLayout(layout)
                dlg.resize(1100, 540)
                dlg.exec()

            tab.warning_label.mouseReleaseEvent = show_error_msg

        if 'ssh ' in tab.server:
            if name in ssh_get_running(tab.server):
                tab.status_label.setText('<font color="green">running</font>')
                return True
            else:
                tab.status_label.setText('<font color="red">inactive</font>')
                return False

        if 'local' in tab.server:
            if tab.process is not None:
                if tab.parent_conn.poll():
                    msg = tab.parent_conn.recv()
                    tab.process.kill()
                    tab.process = None
                    tab.status_label.setText('<font color="red">inactive</font>')
                    return False
                else:
                    tab.status_label.setText('<font color="green">running</font>')
                    return True
            else:
                tab.status_label.setText('<font color="red">inactive</font>')
                return False

    def refresh_view(self, tab):
        return

    def add_common_tab(self, name):

        folder = get_epc_folder(self.folder) + '/' + name + '/'

        tab = self.add_tab(name, stretch=0) #self.Next_Tab(name, stretch=0)
        tab.name = name

        tab.ssm = SimpleStorageManager(folder)
        tab.server = tab.ssm.load_param('server', default=None)

        tab.gene_keys = tab.ssm.load_param('gene_keys', default=None)
        if tab.gene_keys is not None:
            tab.gene_keys = eval(tab.gene_keys)

        tab.process = None
        #tab.python_cmd = tab.ssm.load_param('python_cmd', default='python3')

        if tab.ssm.load_param('executable', default=None) is not None:###################################################################################################

            tab.warning_label = self.tab.add_widget(QLabel(u"\u26A0"))
            tab.warning_label.setVisible(False)

            tab.status_label = self.tab.add_widget(QLabel('Status'))

            def refresh():
                try:
                    print('refresh')
                    if tab.server is not None and 'ssh ' in tab.server:
                        print('get Data')
                        get_Data(tab.name, tab.server, self.folder)
                except:
                    print('no plot data found in', name)

                self.update_status(tab, name)
                self.refresh_view(tab)

            refresh_btn = self.tab.add_widget(QPushButton('refresh state'))
            refresh_btn.clicked.connect(refresh)

            def start_continue():
                if 'ssh ' in tab.server:
                    if not self.update_status(tab, name):
                        ssh_execute_evo(tab.server, name)  # python3
                    else:
                        print('thread already running')

                if 'local' in tab.server:
                    if tab.process is None:
                        tab.parent_conn, child_conn = Pipe()
                        tab.process = Process(target=evolution_thread_worker,
                                              args=(name, self.folder, child_conn))  # python
                        print('start', self.folder+'/'+name+'/execute.py')
                        tab.process.start()
                    else:
                        print('thread already running')

                self.update_status(tab, name)

            start_evo_btn = self.tab.add_widget(QPushButton('start/continue'))
            start_evo_btn.clicked.connect(start_continue)

            def stop():
                if 'ssh ' in tab.server:
                    ssh_stop_evo(tab.server, name)
                    self.update_status(tab, name)
                if 'local' in tab.server:
                    if tab.process is not None:
                        print('Please close terminal manually')
                        tab.process.kill()
                        tab.process.terminate()
                        tab.process.close()
                        tab.process = None

            stop_btn = self.tab.add_widget(QPushButton('stop'))
            stop_btn.clicked.connect(stop)

        #############################################################################################################

        self.tab.add_row(stretch=0)

        self.tab.add_widget(QLabel(tab.server))

        if 'local' in tab.server:
            def on_rename_click():
                self.rename(tab)

            rename_btn = self.tab.add_widget(QPushButton('rename'))
            rename_btn.clicked.connect(on_rename_click)


        if 'ssh' in tab.server:
            def on_connect_click():
                self.show_screen_session(tab.server, name)

            connect_btn = self.tab.add_widget(QPushButton('connect'))
            connect_btn.clicked.connect(on_connect_click)


            def on_localize_click():
                self.localize(tab)

            localize_btn = self.tab.add_widget(QPushButton('localize'))
            localize_btn.clicked.connect(on_localize_click)

        if 'local' in tab.server:
            def archive():
                try:
                    if  tab.server is not None and 'ssh ' in tab.server:
                        ssh_stop_evo(tab.server, name)

                    if  tab.server is not None and 'local' in tab.server:
                        if tab.process is not None:
                            print('Please close terminal manually')
                            tab.process.kill()
                            tab.process.terminate()
                            tab.process.close()
                            tab.process = None
                except:
                    print('evo cannot be stopped')

                zipDir(get_epc_folder(self.folder) + '/' + name, get_epc_folder(self.folder) + '/' + name + '.zip')
                shutil.rmtree(get_epc_folder(self.folder) + '/' + name + '/')
                #self.tabs.removeTab(self.tabs.currentIndex())
                self.remove_tab(tab)

                self.update_archive()

            archive_btn = self.tab.add_widget(QPushButton('archive'))
            archive_btn.clicked.connect(archive)

        def remove():
            if tab.server is not None and 'ssh ' in tab.server:
                ssh_stop_evo(tab.server, name)
                ssh_remove_evo(tab.server, name)

            if tab.server is not None and 'local' in tab.server:
                if tab.process is not None:
                    print('Please close terminal manually')
                    tab.process.kill()
                    tab.process.terminate()
                    tab.process.close()
                    tab.process = None

            shutil.rmtree(get_epc_folder(self.folder) + '/' + name + '/')
            self.remove_tab(tab)
            #self.tabs.removeTab(self.tabs.currentIndex())

        remove_btn = self.tab.add_widget(QPushButton('remove'))
        remove_btn.clicked.connect(remove)


        self.add_additional_tab_elements(tab, name)

        self.On_Tab_Changed(None)

        # TODO: remove
        #smg = StorageManagerGroup(tab.name, data_folder=get_data_folder() + '/' + self.folder + '/' + tab.name + '/Data')
        #params = smg.get_all_params()
        #if 'generation' not in params and 'gen' in params:
        #    for sm in smg.StorageManagerList:
        #        g = sm.load_param('gen')
        #        sm.save_param('generation', g)


    def localize(self, tab):
        name = tab.name
        if tab.server is not None and 'ssh ' in tab.server:
            ssh_stop_evo(tab.server, name)
        transfer_project_back(name, tab.server, main_folder=self.get_folder())
        tab.ssm.save_param('server', 'local')
        ssh_remove_evo(tab.server, name)
        self.remove_tab(tab)
        self.add_common_tab(name)


    def rename(self, tab):
        name = tab.name

        def rename_func(new):
            src = name
            dst = new

            main_folder = get_data_folder() + '/'+self.get_folder()+'/'
            df = main_folder+src+'/Data/StorageManager/'
            data_f = df+src+'/'

            if os.path.exists(data_f):
                for sub_folder in os.listdir(data_f):
                    if src in sub_folder:
                        os.rename(data_f + sub_folder, data_f + sub_folder.replace(src, dst))

                os.rename(df + src, df + dst)

            os.rename(main_folder + src, main_folder + dst)

            self.remove_tab(tab)
            self.add_common_tab(new)

        self.text_input_dialog('type in new name', 'rename', rename_func, 'new_name')

    def show_screen_session(self, server, name):
        ssh = get_ssh_connection(server)
        channel = ssh.invoke_shell(width=500, height=240)

        channel.send(f"screen -r "+name+"\n")
        time.sleep(1)
        output = ''
        # while not channel.exit_status_ready():
        # for i in range(6):
        #    #if channel.recv_ready():
        for i in range(100):
            if channel.recv_ready():
                output += channel.recv(1024).decode('utf-8')
        #print(output)
        channel.close()

        while 'H[J' in output:
            pos = output.find('H[J')
            output=output[pos+4:]


        output = output.replace('min', 'min<br/>').replace('\n', '<br/>')


        dlg = QDialog()
        dlg.setWindowTitle('ssh screen '+name)
        layout = QVBoxLayout()

        edit = QTextEdit(output)
        layout.addWidget(edit)

        def btn_clicked():
            dlg.close()

        btn = QPushButton('close')
        btn.clicked.connect(btn_clicked)

        layout.addWidget(btn)
        dlg.setLayout(layout)
        dlg.resize(1000, 800)
        dlg.exec()



    def on_server_delete(self):
        txt = self.listwidget2.currentText()
        if  txt!= 'local':
            self.listwidget2.removeItem(self.listwidget2.currentIndex())
            ssm = SimpleStorageManager(get_epc_folder(self.folder) + '/')
            #print(ssm.absolute_path)
            servers = eval(ssm.load_param('servers', default="[]"))
            servers.remove(txt)
            ssm.save_param('servers',  servers)


    def on_server_add(self):
        #server_str = self.listwidget2.currentItem().text()
        #if server_str == '+':
        text, ok = QInputDialog.getText(self.main_window, 'Add...', 'Add computing device "ssh user@host" optional arguments:"ssh ... password pw port 123"')
        if ok:
            self.listwidget2.addItem(text)
            ssm = SimpleStorageManager(get_epc_folder(self.folder) + '/')
            print(ssm.absolute_path)
            ssm.save_param('servers', eval(ssm.load_param('servers', default="[]")) + [text])

    def on_file_click(self):
        print(self.listwidget.currentItem().text())

    def on_add_click(self):
        name = self.task_name_edit.text()
        if is_invalid_evo_name(name):
            print(
                'invalid evolution name.\r\n For savety reasons some names and characters are forbidden to avoid the accidental removal of files or folders')
        else:
            item = self.listwidget2.currentText()
            if item is None:
                print('please select a server first')
            else:
                server_str = item
                if os.path.exists(get_data_folder() + '/'+self.folder+'/' + name):
                    self.show_message('Error', 'Folder already exists', 'Ok')
                    print('folder already exists: '+ get_data_folder() + '/'+self.folder+'/' + name)
                elif self.valid_configuration():

                    copy_project(name, server_str, self.folder)

                    folder = get_epc_folder(self.folder) + '/' + name + '/'

                    ssm = SimpleStorageManager(folder)
                    ssm.save_param('server', server_str)
                    ssm.save_param('name', name)

                    self.save_configuration(ssm)

                    evo_file = folder + 'execute.py'
                    if self.create_execution_file(name, folder, evo_file):
                        ssm.save_param('executable', 'True')

                    print('saving config...')

                    if 'ssh' in server_str:
                        print('transfer execute.py to ' + server_str + '...')
                        ssh = get_ssh_connection(server_str)
                        scp = SCPClient(ssh.get_transport())
                        scp.put(evo_file, name + '/execute.py')
                        scp.close()
                        ssh.close()

                    print('adding tab...')
                    self.add_common_tab(name)

if __name__ == '__main__':
    Execution_Manager_UI_Base().show()


import numpy as np

#creates 1D, 2D or 3D generators.
#the grid starts with a low resolution step size (max-min)/5
#and doubles its resolution per axis gradually.
#The longer you let the for loop run, the more detail you will get.

def infinite_resolution_1D(minx, maxx, max_resolution=-1):
    points = np.array([-100000])
    dx = maxx - minx
    r = -1
    while max_resolution==-1 or r<max_resolution:
        r += 1
        resolution = 5*np.power(2, r)
        for a in np.arange(minx, maxx, dx/resolution):
            if np.min(np.abs(points-a)) > 0.00001:
                points = np.concatenate([points, [a]])
                yield a

def infinite_resolution_2D(minx, maxx, miny, maxy, max_resolution=-1):
    points = np.array([[-100000,-100000]])
    dx = maxx - minx
    dy = maxy - miny
    r = -1
    while max_resolution==-1 or r<max_resolution:
        r += 1
        resolution = 5*np.power(2, r)
        for a in np.arange(minx, maxx, dx/resolution):
            for b in np.arange(miny, maxy, dy/resolution):
                new = [a, b]
                if np.min(np.sum(np.abs(points - np.array(new)), axis=1)) > 0.00001:
                    points = np.concatenate([points, [new]])
                    yield new

def infinite_resolution_3D(minx, maxx, miny, maxy, minz, maxz, max_resolution=-1):
    points = np.array([[-100000,-100000,-100000]])
    dx = maxx - minx
    dy = maxy - miny
    dz = maxz - minz
    r = -1
    while max_resolution==-1 or r<max_resolution:
        r += 1
        resolution = 5*np.power(2, r)
        for a in np.arange(minx, maxx, dx/resolution):
            for b in np.arange(miny, maxy, dy/resolution):
                for c in np.arange(minz, maxz, dz/resolution):
                    new = [a, b, c]
                    if np.min(np.sum(np.abs(points - np.array(new)), axis=1)) > 0.00001:
                        points = np.concatenate([points, [new]])
                        yield new


'''
def nth_sqrt(x, n):
    return np.power(x, (1 / n))


def stretched_powerxe(xmin, xmax, x, e):
    scale = (xmax-xmin)/np.power(xmax-xmin, e)
    return xmin+np.power(x-xmin, e)*scale

def stretched_powerxe_center(xmin, xmax, x, e, centerx):
    pcx = np.power(centerx, 1/e)

    sxmin = np.abs(np.power(xmin-pcx, e))
    sxmin=sxmin*(((xmin-pcx)>0)-0.5)*2

    sxmax = np.abs(np.power(xmax-pcx, e))
    sxmin=sxmin*(((xmax-pcx)>0)-0.5)*2

    scale = (xmax-xmin)/(sxmax-sxmin)

    tmp=np.power(x-pcx, e)
    tmp=tmp*(((x-pcx)>0)-0.5)*2

    return xmin+tmp*scale-sxmin*scale

x = np.arange(0.1,3.0,0.1)
import matplotlib.pyplot as plt
#plt.scatter(x,x*0)
#plt.scatter(stretched_log(0.1,3,x,2),x*0+1)
#plt.show()




def infinite_resolution_2D(minx, maxx, miny, maxy):
    result = []
    dx = maxx - minx
    dy = maxy - miny
    r = -1
    for i in range(2):
        r += 1
        resolution = 5*np.power(2, r)
        for a in np.arange(minx, maxx, dx/resolution):
            for b in np.arange(miny, maxy, dy/resolution):
                new = [a, b]
                if new not in result:
                    result.append(new)
    return np.array(result)

d = infinite_resolution_2D(0.1,3.0,0.1,3.0)

plt.scatter(stretched_powerxe_center(0.1, 3, d[:, 0], 2,0),stretched_powerxe_center(0.1, 3, d[:, 1], 2.0, 1.5))
plt.show()


#def steps(start, end, points):
#    result = []
#    if start <= 0 or end<=start:
#        return []

#    exp = nth_sqrt(end/start, points)

#    while start<end:
#        result.append(start)
#        start *= exp
#    result.append(end)
#    return result

'''


from PymoNNto.Exploration.Evolution.Devices.Evolution_Device_Single_Thread import *
from PymoNNto.Exploration.Evolution.Devices.Evolution_Device_SSH import *
from PymoNNto.Exploration.Evolution.Breed_And_Selection_Module import *
from PymoNNto.Exploration.Evolution.Single_Evolution_UI import *
from PymoNNto.Exploration.Evolution.genome_exchange_com import *
import os.path
import time
from PymoNNto.Exploration.Evolution.Evolution_Individual import *


class Evolution:

    def __init__(self, name, slave_file, individual_count=10, mutation=0.4, death_rate=0.5, devices={'single_thread':1}, constraints=[], start_genomes=[], inactive_genome_info={}, breed_and_select_module=Default_Breed_And_Select, additional_evo_params={}):

        self.Breed_And_Select = breed_and_select_module(self, death_rate=death_rate, mutation=mutation, individual_count=individual_count, constraints=constraints, additional_evo_params=additional_evo_params)

        self.genome_exchanger_socket = Genome_Exchanger_Socket()

        self.name = name

        self.slave_file = slave_file

        self.devices = []
        self.start_genomes = start_genomes
        self.inactive_genome_info = inactive_genome_info

        self.id_counter = 0

        if not os.path.isfile(slave_file):
            print('warning slave file not found')

        if start_genomes==[]:
            print('Error no genomes found')

        for key in ['generation', 'score', 'id']:
            for genome in self.start_genomes:
                if key in genome:
                    genome.pop(key)
                    print(key, 'removed form genomes')

        self.gene_keys = []
        if len(self.start_genomes) > 0:
            self.gene_keys = list(self.start_genomes[0].keys())

            individuals = []
            for genome in self.start_genomes:
                individuals.append(Evolution_individual(genome=genome))

            #for genome in self.start_genomes:
            #    if len(self.gene_keys) != len(genome):
            #        print('Error: start genomes do not have the same size')
            #    for gene in genome:
            #        if gene not in self.gene_keys:
            #            print('Error: start genomes have different gene keys')

            #    if not self.Breed_And_Select.is_valid_genome(genome):
            #        print('Error: start genome does not fulfill constraints', genome)

            self.scored_individuals = []
            self.running_individuals = []
            self.non_scored_individuals = self.Breed_And_Select.breed(individuals)

            for individual in self.non_scored_individuals:
                individual.id = self.get_individual_id()

            print('starting individuals:', self.non_scored_individuals)
        else:
            print('Error no start genomes found')

        for device_string, number_of_threads in devices.items():
            self.add_devices(device_string, number_of_threads)

        if is_invalid_evo_name(name):
            print('Error: For savety reasons some names and characters are forbidden to avoid the accidental removal of files or folders')
            self.devices = {}

        for device in self.devices:
            device.initialize()

        self.ui = None

    def get_individual_id(self):
        result = self.id_counter
        self.id_counter += 1
        return result

    def iteration(self):
        for device in self.devices:
            device.main_loop_update()
        #time.sleep(1.0)

    #def part_of_genome(self, small_genome, big_genome):
    #    result = True
    #    for key in small_genome:
    #        if key!='score' and key not in big_genome or small_genome[key] != big_genome[key]:
    #            result = False
    #    return result

    #def find(self, genome_list, genome):
    #    found = None
    #    for g in genome_list:
    #        if self.part_of_genome(g, genome):
    #            found = g
    #    return found

    def new_score_event(self, evo_id, score):
        found = find_individual(evo_id, self.running_individuals)
        if found is not None:
            found.score = score
            self.running_individuals.remove(found)
            self.scored_individuals.append(found)
            self.Breed_And_Select.update_population()
            if self.ui is not None:
                self.ui.update()
        else:
            self.error_event(evo_id, 'processed gene not found in running_individuals | running:' + str(self.running_individuals) + ' scored:' + str(self.scored_individuals))

    #def new_score_event(self, genome):#called by devices
    #    #found = self.find(self.running_individuals, genome)
    #    found = find_individual(genome['id'], self.running_individuals)
    #
    #    if found is not None:
    #        self.running_individuals.remove(found)
    #        found['score'] = genome['score']
    #        self.scored_individuals.append(found)
    #        self.Breed_And_Select.update_population()
    #        print('+', genome)#g
    #
    #        if self.ui is not None:
    #            self.ui.update()
    #
    #    else:
    #        self.error_event(genome, 'processed gene not found in running_individuals | running:' + str(self.running_individuals)+' scored:'+str(self.scored_individuals))

    def error_event(self, evo_id, message, terminate=False):#called by devices
        print('failed', message, 'id=', evo_id)

        if terminate:
            ssm= SimpleStorageManager(get_data_folder()+'/')
            ssm.save_str('error', message+" \r\nid="+str(evo_id))
            print(get_data_folder()+'error.txt write')
        else:
            found = find_individual(evo_id, self.running_individuals)

            if found is not None:
                self.running_individuals.remove(found)
                self.non_scored_individuals.append(found)
                print('move genome from running_individuals back to non_scored_individuals...')
            else:
                print('not able to find failed genome in running_individuals.')

    #def error_event(self, genome, message, terminate=False):#called by devices
    #    print('failed', message, genome)

    #    if terminate:
    #        ssm= SimpleStorageManager(get_data_folder()+'/')
    #        ssm.save_str('error', message+" \r\n"+str(genome))
    #        print(get_data_folder()+'error.txt write')
    #    else:
    #        found = find_individual(genome['id'], self.running_individuals)
    #        #found = None
    #        #for g in self.running_individuals:
    #        #    if self.part_of_genome(g, genome):
    #        #        found = g

    #        if found is not None:
    #            self.running_individuals.remove(found)
    #            self.non_scored_individuals.append(found)
    #            print('move genome from running_individuals back to non_scored_individuals...')
    #        else:
    #            print('not able to find failed genome in running_individuals. Maybe the genome was changed during processing')


    def add_device(self, device_string):
        if device_string == 'single_thread':
            self.devices.append(Evolution_Device_Single_Thread(device_string, self))
        if device_string == 'multi_thread':
            self.devices.append(Evolution_Device_Multi_Thread(device_string, self))
        if 'ssh' in device_string:
            self.devices.append(Evolution_Device_SSH(device_string, self))

    def add_devices(self, device_string, number_of_threads):
        # move folder to remote device if needed
        for _ in range(number_of_threads):
            self.add_device(device_string)

        if len(self.devices) > 0:
            self.devices[-1].initialize_device_group()

    def _run_evo(self, ui=True):
        for device in self.devices:
            device.start()

        if ui:
            self.ui = Single_Evolution_UI(self)
            self.ui.show()
        else:
            self.active = True
            while self.active:

                new_ind = self.genome_exchanger_socket.get_new_individuals()
                if new_ind is not None:
                    if new_ind.has_genes(self.gene_keys):
                        self.non_scored_individuals.append(new_ind)
                        print('socket gene insertion successful', new_ind)
                    else:
                        print('socket gene insertion failed', new_ind)


                for device in self.devices:
                    device.main_loop_update()
                time.sleep(0.01)

        for device in self.devices:
            device.stop()

    def load_state(self):
        population = []

        smg = StorageManagerGroup(self.name)

        smg.sort_by('generation')
        gens = smg.get_param_list('generation', remove_None=True)
        if len(gens)>0:
            last_full_gen = np.max(gens) - 1
            for sm in smg['generation==' + str(last_full_gen)]:
                individual = Evolution_individual(genome={})

                valid = True
                for gene in self.gene_keys:
                    individual.genome[gene] = sm.load_param(gene)
                    if individual.genome[gene] is None:
                        valid = False

                individual.score = sm.load_param('score')

                if individual.score is None:
                    valid = False

                if valid:
                    population.append(individual)
                else:
                    print('invalid individual!')

            ids = smg.get_param_list('id', remove_None=True)
            max_id = np.max(ids)
            self.id_counter = max_id + 1

            if len(population) != 0:
                self.scored_individuals = population
                self.running_individuals = []
                self.non_scored_individuals = []
                self.Breed_And_Select.generation = last_full_gen

            print('Continuing', self.name, 'at generation', last_full_gen, 'with genomes', self.scored_individuals)

        self.Breed_And_Select.update_population()

    def continue_evolution(self, ui=True):
        self.load_state()
        self._run_evo(ui)

    def start(self, ui=True):
        folder = get_data_folder()+'/StorageManager/'+self.name
        if not os.path.isdir(folder):
            self._run_evo(ui)
            return True
        else:
            print('Warning:', folder, 'already exists. Remove folder or try continue_evolution() instead of start()')
            return False

    def stop(self):
        for device in self.devices:
            device.stop()

    #def add_name_gen_id_inactive_to_genome(self, genome):
    #    #result = genome.copy()
    #    result['evo_name'] = self.name
    #    #result['generation'] = self.Breed_And_Select.generation
    #    #result['id'] = self.id_counter
    #    self.id_counter += 1
    #    #result.update(self.inactive_genome_info)
    #    return result

    def get_next_individual(self):
        result = None
        if len(self.non_scored_individuals) > 0:
            result = self.non_scored_individuals[0]
            self.non_scored_individuals.pop(0)
            self.running_individuals.append(result)

        #if result is not None:
        #    result = self.add_name_gen_id_inactive_to_genome(result)

        return result

    #add evo_name and static genome to pipeline!!!


from PymoNNto.Exploration.Evolution.Devices.Evolution_Device import *

class Evolution_Device_Single_Thread(Evolution_Device):

    def main_loop_update(self):
        current_individual = self.parent.get_next_individual()

        if current_individual is not None:

            #if os.path.isfile(self.parent.slave_file):
            #print('exec', current_individual.id)
            execute_local_file(self.parent.slave_file, self.parent.name , current_individual.id, self.parent.Breed_And_Select.generation, current_individual.genome, self.parent.inactive_genome_info)
            #else:
            #    print('file not found', self.parent.slave_file)

            self.score_processing(current_individual.id)

        else:
            print('no genes found to process')



from PymoNNto.Exploration.Network_UI.TabBase import *

class weight_tab(TabBase):

    def __init__(self, title='Weights', weight_attrs=['W']):
        super().__init__(title)
        self.weight_attrs = weight_attrs

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        self.weight_tab = Network_UI.add_tab(title=self.title)
        self.main_plot = Network_UI.tab.add_plot()

        Network_UI.tab.add_row()

        self.select_mode = QComboBox()
        self.select_mode.addItems(['selected neuron', 'all'])
        self.select_mode.setCurrentIndex(0)
        Network_UI.tab.add_widget(self.select_mode, stretch=1)

        self.select_aff_eff = QComboBox()
        self.select_aff_eff.addItems(['afferent', 'efferent'])
        self.select_aff_eff.setCurrentIndex(0)
        Network_UI.tab.add_widget(self.select_aff_eff, stretch=1)

        self.select_min_box = QComboBox()
        self.select_min_box.addItems(['global min', 'single min', '0'])
        self.select_min_box.setCurrentIndex(1)
        Network_UI.tab.add_widget(self.select_min_box, stretch=1)

        self.select_max_box = QComboBox()
        self.select_max_box.addItems(['global max', 'single max', '1'])
        self.select_max_box.setCurrentIndex(1)
        Network_UI.tab.add_widget(self.select_max_box, stretch=1)


    def update(self, Network_UI):
        if self.weight_tab.isVisible():
            group = Network_UI.selected_neuron_group()
            selected = Network_UI.selected_neuron_id()

            self.main_plot.clear()

            aff=self.select_aff_eff.currentText() == 'afferent'
            single = self.select_mode.currentText() != 'all'
            max_mode = self.select_max_box.currentText()
            min_mode = self.select_min_box.currentText()

            c = 0
            syns = []
            syn_tags=[]
            max_s = []
            min_s = []

            if aff:
                for s in group.synapses(afferent):
                    w = s.ignore_transpose_mode(s.W)

                    if single:
                        data = np.rot90(w[selected].reshape(s.src.height, s.src.width), 3)
                    else:
                        data = w
                    syns.append(data)
                    syn_tags.append(s.tags)
            else:
                for s in group.synapses(efferent):#efferent_synapses['All']:
                    w = s.ignore_transpose_mode(s.W)

                    if single:
                        data = np.rot90(w[:, selected].reshape(s.dst.height, s.dst.width), 3)
                    else:
                        data = w
                    syns.append(data)
                    syn_tags.append(s.tags)

            for data in syns:
                max_s.append(np.max(data))
                min_s.append(np.min(data))

            max_s = np.array(max_s)
            min_s = np.array(min_s)

            if len(max_s) > 0:
                if max_mode == 'global max':
                    max_s[:] = np.max(max_s)
                elif max_mode == 'single max':
                    max_s[:] = max_s[:]
                else:
                    max_s[:] = float(max_mode)

                if min_mode == 'global min':
                    min_s[:] = np.max(min_s)
                elif min_mode == 'single min':
                    min_s[:] = min_s[:]
                else:
                    min_s[:] = float(min_mode)

            for data, ma, mi, tags in zip(syns, max_s, min_s, syn_tags):
                image = self.main_plot.add_image()
                image.setImage(data, levels=(mi, ma))
                text = self.main_plot.add_text(str(tags))

                text.setPos(0, c+10)
                image.setRect(0, c-data.shape[1], data.shape[0], data.shape[1])
                c -= data.shape[1] + 20



from PymoNNto.Exploration.Evolution.communication import *
from PymoNNto.Exploration.Evolution.Evolution_Individual import *
import ast



class Genome_Exchanger_Socket(Socket_Listener):

    def __init__(self):
        super().__init__()
        self.individuals = []

    def get_new_individuals(self):
        if len(self.individuals)>0:
            return self.individuals.pop(0)
        return None

    def handle_message(self, data):
        lines = data.split("\r\n")
        if lines[0] == "insert":
            self.individuals = []
            valid = 0
            for l in lines[1:]:
                try:
                    genome = ast.literal_eval(l)
                    self.individuals.insert(0, Evolution_individual(genome))
                    valid += 1
                except:
                    print("received invalid genome", l)
            return str(valid) + " genomes inserted"
        return "no known command"


from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.Exploration.Visualization.Visualization_Helper import *
import os
from io import StringIO
from contextlib import redirect_stdout

from PymoNNto.Exploration.Network_UI.Basic_Tabs.Helper.syntax import *

class code_execution_tab(TabBase):

    sample_code1='''import matplotlib.pyplot as plt

synapse_tag='GLU'
synapse_var='W'

combined = get_combined_syn_mats(n.synapses(afferent, synapse_tag), None,synapse_var)

for mat in combined.values():

	plt.imshow(mat, vmin=0, vmax=0.005)#cmap='gray', 
	plt.show()
'''

    sample_code2='''num=0

for syn in n.synapses(afferent, 'GLU'):

	num+=np.sum(syn.W>0.001)

print(num)
'''

    base_code = '''###Quick Access Variables###
    
# network | net
# neuron_group | neurons | ng | n
# Network_UI | ui | UI
# storage_manager | storage | sm

###Python Code###


'''

    def initialize(self, Network_UI):
        self.Network_UI=Network_UI

        self.code_execution_tab = Network_UI.add_tab(title='Code Exec.') #Network_UI.Next_Tab()
        self.comboBox = Network_UI.tab.add_widget(QComboBox())

        #self.CB_line_edit = QLineEdit()
        #self.comboBox.setLineEdit(self.CB_line_edit)

        self.txt_py_dict={}

        self.txt_py_dict[''] = self.base_code
        self.txt_py_dict["show synapse matrices"] = self.sample_code1
        self.txt_py_dict["count synapses"] = self.sample_code2

        data_folder = get_data_folder(create_when_not_found=False)+'/scripts'

        if not os.path.exists(data_folder):
            os.mkdir(data_folder)

        for d in os.listdir(data_folder):
            try:
                if '.py' in d:
                    with open(data_folder+'/'+d) as f:
                        self.txt_py_dict[d.replace('.py','')] = f.read()
            except:
                print("error in script", d)


        for txt in self.txt_py_dict:
            self.comboBox.addItem(txt)


        def on_cb_click(event):
            txt=self.comboBox.itemText(self.comboBox.currentIndex())
            self.code_field.setPlainText(self.txt_py_dict[txt])#setText
            #self.code_field.setStyleSheet("QTextEdit { background-color: rgb(255, 255, 255); }")
            self.comboBox.setStyleSheet("background-color: rgb(255, 255, 255);")

        def on_code_changed():
            #self.code_field.setStyleSheet("QTextEdit { background-color: rgb(255, 210, 210); }")
            self.comboBox.setStyleSheet("background-color : rgb(255, 210, 210);")


        self.comboBox.currentIndexChanged.connect(on_cb_click)

        Network_UI.tab.add_row()

        #self.code_field=Network_UI.tab.add_widget(QTextEdit())
        #self.code_field.setAcceptRichText(False)
        #self.code_field.setText(self.base_code)
        #self.code_field.textChanged.connect(on_code_changed)

        net = Network_UI.network
        network = net
        n = Network_UI.selected_neuron_group()
        ng = n
        neurons = n
        neuron_group = n

        self.code_field = Network_UI.tab.add_widget(QPlainTextEdit())
        special_keywords = ['Network_UI', 'net', 'n', 'ng', 'network', 'neurons', 'neuron_group']
        self.code_field.highlight = PythonHighlighter(self.code_field.document(), special_keywords)
        #self.code_field.setAcceptRichText(False)
        self.code_field.setStyleSheet("QPlainTextEdit { background-color: rgb(43, 43, 43); }")
        self.code_field.setPlainText(self.base_code)#setText
        self.code_field.textChanged.connect(on_code_changed)

        Network_UI.tab.add_row(stretch=1)

        self.exec_btn=Network_UI.tab.add_widget(QPushButton('Execute'))

        def exec_btn_click(event):
            ui = Network_UI
            UI = ui

            sm = Network_UI.storage_manager
            storage = sm
            storage_manager = sm

            net = Network_UI.network
            network = net

            n = Network_UI.selected_neuron_group()
            ng = n
            neurons = n
            neuron_group = n

            if self.show_output_window.isChecked():
                f = StringIO()
                #try:
                with redirect_stdout(f):
                    exec(self.code_field.toPlainText())
                #except Exception:
                #        pass
                s = f.getvalue()

                layout = QVBoxLayout()
                pte = QPlainTextEdit()
                pte.setPlainText(s)
                pte.setReadOnly(True)
                layout.addWidget(pte)

                dlg = QDialog()
                dlg.setWindowTitle('code stdout (prints)')
                dlg.setLayout(layout)
                dlg.resize(600, 400)
                dlg.exec()
            else:
                exec(self.code_field.toPlainText())


            print('code block executed successfuly!')

            Network_UI.add_event('code execution')

        self.exec_btn.clicked.connect(exec_btn_click)


        self.show_output_window = Network_UI.tab.add_widget(QCheckBox('Show stdout in window'))

        self.compiled = None
        self.compiled_script_txt = ''
        self.timestep_execute_cb = Network_UI.tab.add_widget(QCheckBox('Execute every timestep'))


        self.save_btn_click = Network_UI.tab.add_widget(QPushButton('Save Script'))

        def save(input_txt):
            data_folder = get_data_folder(create_when_not_found=False)+'/scripts'
            name=data_folder + '/'+input_txt+'.py'
            with open(name, "w") as f:
                f.write(self.code_field.toPlainText())

            n=input_txt
            found = n in self.txt_py_dict

            self.txt_py_dict[n] = self.code_field.toPlainText()

            if not found:
                self.comboBox.addItem(n)
                self.comboBox.setCurrentIndex(len(self.txt_py_dict)-1)

            self.code_field.setStyleSheet("QTextEdit { background-color: rgb(255, 255, 255); }")

        def save_btn_click(event):
            Network_UI.text_input_dialog("Please name your script", 'save', save, self.comboBox.itemText(self.comboBox.currentIndex()).replace('.py',''))

        self.save_btn_click.clicked.connect(save_btn_click)



    def update(self, Network_UI):

        if self.timestep_execute_cb.isChecked():
            self.show_output_window.setChecked(False)

            net = Network_UI.network
            network = net
            n = Network_UI.selected_neuron_group()
            ng = n
            neurons = n
            neuron_group = n

            code = self.code_field.toPlainText()

            if self.compiled is None or self.compiled_script_txt != code:
                self.compiled = compile(code, '<string>', 'exec')
                self.compiled_script_txt = code

            exec(self.compiled)




from PymoNNto import *

class refractory(Behavior):

    def initialize(self, neurons):
        neurons.refractory_counter = neurons.vector()
        self.decayfactor = self.parameter('decayfactor', 0.5, neurons)
        self.strengthfactor = self.parameter('strengthfactor', 1.0, neurons)

    def iteration(self, neurons):
        neurons.refractory_counter *= self.decayfactor
        neurons.refractory_counter += neurons.output

        neurons.activity -= neurons.refractory_counter * self.strengthfactor


from PymoNNto.NetworkCore.Base_Tagable_Object import *
from PymoNNto.Exploration.Evolution.Interface_Functions import *



class Behavior(TaggableObjectBase):

    initialize_on_init = False
    initialize_last = False
    attached_UI_Tabs = []


    def initialize(self, neurons):
        return

    def iteration(self, neurons):
        return


    def __init__(self, *args, **kwargs):
        self.init_kwargs = kwargs
        for i, arg in enumerate(args):
           self.init_kwargs['arg_'+str(i)]=arg
        self.used_attr_keys = []
        self.behavior_enabled = self.parameter('behavior_enabled', True, None)
        super().__init__(tag=self.parameter('tag', None, None))
        self.empty_iteration_function = self.is_empty_iteration_function()

        if hasattr(self, 'set_variables'):
            raise Exception('"set_variables" is deprecated. Please rename function to "initialize"')
        if hasattr(self, 'new_iteration'):
            raise Exception('"new_iteration" is deprecated. Please rename function to "iteration"')


    def set_gene_variables(self):

        for variable_key in self.init_kwargs:
            while type(self.init_kwargs[variable_key]) is str and '[' in self.init_kwargs[variable_key] and ']' in self.init_kwargs[variable_key]:
                s = self.init_kwargs[variable_key]

                start = s.index('[')
                end = s.index(']')

                content = s[start + 1: end]

                if '#' in content:
                    parts = content.split('#')
                    default_value = eval(parts[0])#float(parts[0])
                    gene_key = parts[1]
                else:
                    gene_key = content
                    default_value = None

                gene_value = get_gene(gene_key, default_value)

                self.init_kwargs[variable_key] = s[:start] + '{:.15f}'.format(gene_value).rstrip('0').rstrip('.') + s[end + 1:]

    def __str__(self):
        result = self.__class__.__name__+'('
        for k in self.init_kwargs:
            result += str(k) + '=' + str(self.init_kwargs[k])+','
        result += ')'
        return result

    def evaluate_diversity_string(self, ds, neurons_or_synapses):

        if 'same(' in ds and ds[-1] == ')':
            params = ds[5:-1].replace(' ', '').split(',')
            if len(params) == 2:
                return getattr(neurons_or_synapses[params[0], 0], params[1])

        plot = False
        if ';plot' in ds:
            ds = ds.replace(';plot', '')
            plot = True

        result = ds

        if '(' in ds and ')' in ds:#is function
            if type(neurons_or_synapses).__name__ == "NeuronGroup":
                result = neurons_or_synapses.vector(ds)

            if type(neurons_or_synapses).__name__ == "SynapseGroup":
                result = neurons_or_synapses.matrix(ds)

        if plot:
            if type(result) == np.ndarray:
                import matplotlib.pyplot as plt
                plt.hist(result, bins=30)
                plt.show()

        return result


    def set_parameters_as_variables(self, object):
        for key in self.init_kwargs:
            setattr(object, key, self.parameter(key, None, neurons=object))

    set_init_attrs_as_variables = set_parameters_as_variables

    def check_unused_attrs(self):
        for key in self.init_kwargs:
            if not key in self.used_attr_keys:
                print('Warning: "'+key+'" not used in initialize of '+str(self)+' behavior! Make sure that "'+key+'" is spelled correctly and parameter('+key+',...) is called in initialize. Valid attributes are:'+str(self.used_attr_keys))

    def parameter(self, key, default=None, neurons=None, do_not_diversify=False, search_other_behaviors=False, required=False):

        if required and not key in self.init_kwargs:
            print('Warning:',key,'has to be specified for the behavior to run properly.', self)

        self.used_attr_keys.append(key)

        result = self.init_kwargs.get(key, default)

        if key not in self.init_kwargs and neurons is not None and search_other_behaviors:
            for b in neurons.behaviors:
                if key in b.init_kwargs:
                    result = b.init_kwargs.get(key, result)

        if not do_not_diversify and type(result) is str and neurons is not None:
            result = self.evaluate_diversity_string(result, neurons)

        if type(result) is str and default is not None:
            if '%' in result and is_number(result.replace('%', '')):
                result = str(float(result.replace('%', '')) / 100.0)

            try:
                result = type(default)(result)#cast
            except:
                result = float(result)#to prevent int error during evolution


        return result

    get_init_attr = parameter
    kwargs = parameter
    args = parameter
    parameter = parameter
    argument = parameter

    def visualize_module(self, vmi=None, vmo=None, vma=None):
        from PymoNNto.Exploration.Visualization import Module_visualizer as drawer
        self.visualization_module_inputs = vmi
        self.visualization_module_outputs = vmo
        self.visualization_module_attributes = vma
        md = drawer.module_drawer()
        md.add_module(self)
        md.show()

    #helper function for UI

    def get_UI_Tabs(self):
        return []

    def get_UI_Preview_Plots(self):
        #Examples:
        #[[np.sin(x) for x in range(100)]] # only y
        #[[list(range(100)),[np.sin(x) for x in range(100)]]] x and y
        #[[np.sin(x) for x in range(100)],[np.sin(x) for x in range(100)]] # two plots only y
        #[np.random.rand(291, 291, 3)] image
        #[np.random.rand(291, 291, 3), [np.sin(x) for x in range(100)]] image and plot
        return None

    def is_empty_iteration_function(self):
        f = self.iteration

        #Returns true if f is an empty function.
        def empty_func():
            pass

        def empty_func_with_docstring():
            """Empty function with docstring."""
            pass

        def constants(f):
            #Return a tuple containing all the constants of a function without: * docstring
            return tuple(x for x in f.__code__.co_consts if x != f.__doc__)

        return (f.__code__.co_code == empty_func.__code__.co_code and constants(f) == constants(empty_func)) or \
               (f.__code__.co_code == empty_func_with_docstring.__code__.co_code and constants(f) == constants(empty_func_with_docstring))


class Behaviour(Behavior):
    def __init__(self, *args, **kwargs):
        raise Exception("Warning: 'Behaviour' is depreciated. Please use 'Behavior' instead!")




from PymoNNto.Exploration.StorageManager.StorageManager import *
import numpy as np

#convenience functions:

#add root project folder to import path
add_project_root_path()

#print numpy arrays with commas for easier copy paste
np.set_string_function(lambda x: repr(x), repr=False)

from PymoNNto.NetworkCore.Network import *
from PymoNNto.NetworkCore.Behavior import *
from PymoNNto.NetworkCore.Neuron_Group import *
from PymoNNto.NetworkCore.Synapse_Group import *
from PymoNNto.NetworkCore.Analysis_Module import *
from PymoNNto.NetworkBehavior.Structure.Structure import *
from PymoNNto.NetworkBehavior.Structure.Partition import *
from PymoNNto.NetworkBehavior.Structure.Receptive_Fields import *
from PymoNNto.NetworkBehavior.Recorder.Recorder import *
from PymoNNto.Exploration.HelperFunctions import *





from PymoNNto import *
from PymoNNto.Exploration.Evolution.Interface_Functions import *
import numpy as np
import matplotlib.pyplot as plt

class Counter(Behavior):
    def initialize(self, neurons):
        self.inc = self.parameter('inc', 1.0)
        neurons.count = np.zeros(neurons.size)#neurons.vector()
        plt.plot([1,2,3])
        plt.show()

    def iteration(self, neurons):
        neurons.count += self.inc


My_Network = Network()
My_Neurons = NeuronGroup(net=My_Network, tag='my_neurons', size=100, behavior={
    1: Counter(inc='[2.0#a]'),
    2: Recorder('count')
})
My_Synapses = SynapseGroup(net=My_Network, src=My_Neurons, dst=My_Neurons, tag='GLUTAMATE')
sm = StorageManager('test', random_nr=False, print_msg=False)
My_Network.initialize(storage_manager=sm)

My_Network.simulate_iterations(100)

score = np.mean(My_Neurons.count)

set_score(score)





def find_individual(id, ind_list):
    for ind in ind_list:
        if ind.id == id:
            return ind



class Evolution_individual:

    def __init__(self, genome, id=None, score=None, mate_chance=None):
        self.id = id
        self.genome = genome
        self.score = score
        self.mate_chance = mate_chance

    def get_copy(self):
        return Evolution_individual(genome = self.genome.copy(), mate_chance = self.mate_chance)

    def is_valid_genome(self, constraints):
        if self.genome is None:
            return False
        else:
            result = True
            for gene_key, gene_value in self.genome.items():
                locals()[gene_key] = gene_value

            for constraint in constraints:
                result = result and eval(constraint)#todo: try except and error message

            return result

        return False

    def has_genes(self, gene_keys):
        for key in gene_keys:
            if not key in self.genome:
                return False
        return True

    def has_compatible_genes(self, individual):
        if len(individual.genome) != len(self.genome): #different gene length
            return False

        for key in individual.genome: #different genes
            if not key in self.genome:
                return False

        return True

    def __str__(self):
        return str(self.genome)

    def __repr__(self):
        return self.__str__()


from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.NetworkBehavior.Recorder.Recorder import *

class individual_weight_tab(TabBase):

    def __init__(self, title='Ind Weights', weight_attr='W'):
        super().__init__(title)
        self.weight_attr = weight_attr

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        self.individual_weight_tab = Network_UI.Next_Tab(self.title)

        self.plot = Network_UI.Add_plot('Neuron Weights', stretch = 8)

        self.rec_name=self.title + '_rec'
        self.Network_UI=Network_UI

        self.current_ng=None
        self.neuron_index=-1
        self.my_syn_rec=None

        self.key=''

        self.listWidget = Network_UI.Add_element(QListWidget())

        Network_UI.Next_H_Block()

        self.min_w_slider = QSlider(1)  # QtCore.Horizontal
        self.min_w_slider.setMinimum(-1)
        self.min_w_slider.setMaximum(100)
        self.min_w_slider.setSliderPosition(-1)
        self.min_w_slider.setToolTip('slide to cut away x% of smallest weights')
        self.min_w_slider.mouseReleaseEvent = Network_UI.static_update_func
        Network_UI.Add_element(self.min_w_slider)


    def start_rec(self, group):

        if self.current_ng is not None and (self.neuron_index != self.Network_UI.selected_neuron_id() or self.current_ng != group):

            for s in self.current_ng.synapses(afferent):
                s.UI_recorder = None
                s.cb = None
                self.Network_UI.network.remove_behaviors_from_object(s, tags=[self.key])

            self.listWidget.clear()
            self.plot.clear()

            self.current_ng = None
            self.neuron_index = -1
            print('removed')

        if self.current_ng is None:

            self.current_ng = group
            self.neuron_index = self.Network_UI.selected_neuron_id()

            print('started')
            self.plot.clear()

            self.key = self.weight_attr+'['+str(self.neuron_index)+']'

            self.checkboxes={}
            self.listWidget.clear()

            for s in group.synapses(afferent):
                s.UI_recorder = Recorder([self.key, 'iteration'], tag=self.rec_name, gapwidth=10, save_as_numpy=True)
                s.add_behavior(10001, s.UI_recorder)
                #self.Network_UI.network.add_behaviors_to_object({10001: s.UI_recorder}, s)

                s.UI_curves = []
                syn_data = eval(self.key)

                s.cb = QListWidgetItem(str(s.tags[0]))
                s.cb.setCheckState(2)
                self.listWidget.addItem(s.cb)

                for i in range(len(syn_data)):
                    if (type(s.enabled) is bool and s.enabled) or s.enabled[self.neuron_index, i]:
                        color = (s.src.color[0], s.src.color[1], s.src.color[2], 255)
                        curve = pg.PlotCurveItem([], pen=color)  # pen=colors[i%len(colors)]
                        curve.index = i
                        self.plot.addItem(curve)
                        s.UI_curves.append(curve)


    def update(self, Network_UI):

        if self.individual_weight_tab.isVisible():

            group = Network_UI.selected_neuron_group()
            self.start_rec(group)

            for s in self.current_ng.synapses(afferent):
                if not s.UI_recorder.behavior_enabled:
                    s.UI_recorder.behavior_enabled = True
                    print('recorder started')
                if s.cb.checkState() == 2:
                    if s.UI_recorder.is_new_data_available():
                        x_data = s.UI_recorder['iteration', 0]
                        y_data = s.UI_recorder[self.key, 0]

                        if len(x_data) > 0 and len(y_data) > 0:
                            for curve in s.UI_curves:
                                curve.setData(x_data, y_data[:, curve.index])
        else:
            if self.current_ng is not None:
                for s in self.current_ng.synapses(afferent):
                    if s.UI_recorder.behavior_enabled:
                        s.UI_recorder.behavior_enabled=False
                        print('recorder paused')



from PymoNNto.Exploration.Network_UI.TabBase import *

class sidebar_module_quick_access(TabBase):

    def __init__(self, module_tags=[]):
        self.module_tags = module_tags

    def module_on_off(self, cb, Network_UI):
        module_tag=cb.module_tag
        print(module_tag)
        Network_UI.network.set_behaviors(module_tag, cb.isChecked())
        Network_UI.add_event(module_tag + ' (' + str(cb.isChecked()) + ')')

    def add_cb(self, module_tag, Network_UI):

        def module_on_off(event):
            print(module_tag)
            Network_UI.network.set_behaviors(module_tag, cb.isChecked())
            Network_UI.add_event(module_tag + ' (' + str(cb.isChecked()) + ')')

        cb = QCheckBox()
        cb.module_tag = module_tag
        cb.setText(module_tag)
        cb.setChecked(Network_UI.network[module_tag, 0].behavior_enabled)
        cb.stateChanged.connect(module_on_off)
        Network_UI.sidebar.add_widget(cb)

    def initialize(self, Network_UI):

        for module_tag in self.module_tags:
            if Network_UI.network.has_module(module_tag):
                self.add_cb(module_tag, Network_UI)



import numpy as np

class TaggableObjectBase:

    def __init__(self, tag):
        self.tags = []
        self.clear_cache()
        self._mat_eval_dict = {}
        if tag is not None:
            self.add_tag(tag)
        self.add_tag(self.__class__.__name__)
        self._searching = False

    @property
    def tag(self):
        return self.tags[0]

    def has_module(self, tag):
        return self[tag, 0] is not None

    def find_objects(self, key):  # override for deeper search
        result = []
        return result

    def clear_cache(self):
        self.tag_shortcuts = {}

    def set_tag_attrs(self, tag, attr, value):
        for obj in self[tag]:
            setattr(obj, attr, value)

    def call_tag_functions(self, tag, attr, **args):  # todo: test
        for obj in self[tag]:
            getattr(obj, attr)(args)

    def get(self, key, first=True, numpy=True):
        return self[key]  # todo

    def __getitem__(self, key):
        np_array_conversion = isinstance(key, tuple) and 'np' in key

        if isinstance(key, tuple):
            k = key[0]
            index = key[1]
        else:
            k = key
            index = None

        # cache
        if key in self.tag_shortcuts and type(self.tag_shortcuts[key]) is not np.ndarray:
            if np_array_conversion:
                result = np.array(self.tag_shortcuts[key])
            else:
                result = self.tag_shortcuts[key]

            if index is not None:
                if len(result) > 0:
                    return result[index]
                else:
                    return None
            else:
                return result

        # normal search
        result = []
        if k in self.tags or (k is type and isinstance(self, k)):
            result = [self]

        result += self.find_objects(k)

        if k not in self.tag_shortcuts:
            self.tag_shortcuts[k] = result

        if index is not None:
            if len(result) > 0:
                result = result[index]
            else:
                result = None

        if np_array_conversion:
            return np.array(result)
        else:
            return result

    def add_tag(self, tag):
        for subtag in tag.split(','):
            self.tags.append(subtag)
        return self



import socket
import threading

PORT = 45350

#HOST = "127.0.0.1"  # Standard loopback interface address (localhost)


def socket_send(ip, data_str):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.connect((ip, PORT))
        s.sendall(str.encode(data_str))#b"Hello, world"
        #print("client sent", data_str)
        data = s.recv(1024)
        data = data.decode('utf-8')  # to str
        #print("client received", data)
    return data#response


class Socket_Listener:

    def __init__(self, start_listening=True, message_handler_func=None):# example: handle_message(data_str)=>answer_str
        if message_handler_func is not None:
            self.message_handler_func = message_handler_func
        else:
            self.message_handler_func = self.handle_message
        if start_listening:
            self.start()

    def handle_message(self, data):
        return "None"

    def _listen(self):
        #try:
        if True:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)#
                s.bind(("0.0.0.0", PORT))#socket.gethostname()
                s.listen()
                while True:
                    conn, addr = s.accept()
                    with conn:
                        while True:
                            data = conn.recv(1024)
                            if not data:
                                break
                            data = data.decode('utf-8')  # to str
                            print('received', data)

                            response_str=self.message_handler_func(data)

                            print('send', response_str)
                            conn.sendall(str.encode(response_str))
        #except:
        #    print('socket can not be initialized')

    def start(self):
        self.t=threading.Thread(target=self._listen, args=())
        self.t.start()



#sl = Socket_Listener()
#print(socket_send("localhost", "insert\r\n{'a':4, 'b':6, 'c':1}\r\n{'a':6, 'b':6, 'c':6}"))




from PymoNNto.Exploration.Network_UI.TabBase import *
import scipy.fftpack

class fourier_tab(TabBase):

    def __init__(self, parameter, title='Fourier', timesteps=1000):
        super().__init__(title)
        self.parameter = parameter
        self.timesteps=timesteps

    def add_recorder_variables(self, neuron_group, Network_UI):
        try:
            Network_UI.add_recording_variable(neuron_group, 'np.mean('+self.parameter+')', timesteps=self.timesteps)
        except:
            print(self.parameter, 'cannot be added to recorder')

    def initialize(self, Network_UI):

        self.fourier_tab = Network_UI.add_tab(title=self.title)

        exc_plt = Network_UI.tab.add_plot(title='Selected Group Frequencies', x_label='Frequency [Hz]', y_label='Amplitude')
        curves = exc_plt.add_curves(colors=[(0, 0, 0)], legend=False, number_of_curves=6)

        self.exc_fft_curve, self.exc_alpha, self.exc_beta, self.exc_theta, self.exc_delta, self.exc_gamma = curves

        self.exc_fft_curve.setData([0, 0], [0, 0.1])

        text = pg.TextItem(text='Delta', anchor=(0.5, 0))
        text.setPos(2, 0)
        exc_plt.addItem(text)

        text = pg.TextItem(text='Theta', anchor=(0.5, 0))
        text.setPos(6, 0)
        exc_plt.addItem(text)

        text = pg.TextItem(text='Alpha', anchor=(0.5, 0))
        text.setPos(11, 0)
        exc_plt.addItem(text)

        text = pg.TextItem(text='Beta', anchor=(0.5, 0))
        text.setPos(22, 0)
        exc_plt.addItem(text)

        text = pg.TextItem(text='Gamma', anchor=(0.5, 0))#(0, 0)
        text.setPos(40, 0)
        exc_plt.addItem(text)

        Network_UI.tab.add_row()

        Network_UI.tab.add_widget(QLabel('Min:'))

        self.cut_slider = QSlider(1)
        self.cut_slider.setMinimum(0)
        self.cut_slider.setMaximum(10)
        self.cut_slider.setSliderPosition(1)
        Network_UI.tab.add_widget(self.cut_slider)

        Network_UI.tab.add_row()

        self.mspc_label=QLabel('ms/cycle:')
        Network_UI.tab.add_widget(self.mspc_label)

        self.ms_per_cycle_slider = QSlider(1)
        self.ms_per_cycle_slider.setMinimum(1)
        self.ms_per_cycle_slider.setMaximum(100)
        self.ms_per_cycle_slider.setSliderPosition(10)
        Network_UI.tab.add_widget(self.ms_per_cycle_slider)

        Network_UI.tab.add_row()

        Network_UI.tab.add_widget(QLabel('Smoothing:'))

        self.smooth_slider = QSlider(1)
        self.smooth_slider.setMinimum(0)
        self.smooth_slider.setMaximum(50)
        self.smooth_slider.setSliderPosition(0)
        Network_UI.tab.add_widget(self.smooth_slider)

    def smooth(self, data):
        smoothed = data.copy()
        if len(data) > 1:
            for _ in range(int(self.smooth_slider.sliderPosition())):
                for i in range(len(data)):
                    if i == 0:
                        smoothed[i] = (smoothed[i] + smoothed[i + 1] + data[i]) / 3.0
                    elif i == len(data) - 1:
                        smoothed[i] = (smoothed[i] + smoothed[i - 1] + data[i]) / 3.0
                    else:
                        smoothed[i] = (smoothed[i] + smoothed[i - 1] + smoothed[i + 1]) / 3.0
        return smoothed

    def update(self, Network_UI):
        if self.fourier_tab.isVisible():

            group = Network_UI.selected_neuron_group()

            try:
                cut = int(self.cut_slider.sliderPosition())
                ms_per_cycle = int(self.ms_per_cycle_slider.sliderPosition())
                self.mspc_label.setText('ms/cycle: {}'.format(ms_per_cycle))

                exc_act = group['np.mean('+self.parameter+')', 0, 'np'][-self.timesteps:]
                N = len(exc_act)
                T = ms_per_cycle / 1000
                yf = scipy.fftpack.fft(exc_act)
                xf = np.linspace(int(1.0 - N / 1000), int(1.0 / (2.0 * T)), int(N / 2))[cut:]
                real = (2.0 / N * np.abs(yf[:N // 2]))[cut:]
                self.exc_fft_curve.setData(xf, self.smooth(real))

                max_y = np.max(real) + 0.001
                max_x = np.max(xf)

                self.exc_delta.setData([0.1, 3.9], [0, 0], fillLevel=max_y, brush=(50, 250, 50, 50), pen=(50, 250, 50, 50))
                self.exc_theta.setData([min(max_x, 4), min(max_x, 7.9)], [0, 0], fillLevel=max_y, brush=(100, 200, 50, 50), pen=(50, 250, 50, 50))
                self.exc_alpha.setData([min(max_x, 8), min(max_x, 13.9)], [0, 0], fillLevel=max_y, brush=(150, 150, 50, 50), pen=(50, 250, 50, 50))
                self.exc_beta.setData([min(max_x, 14), min(max_x, 29.9)], [0, 0], fillLevel=max_y, brush=(200, 100, 50, 50), pen=(50, 250, 50, 50))
                self.exc_gamma.setData([min(max_x, 30), min(max_x, 100)], [0, 0], fillLevel=max_y, brush=(250, 50, 50, 50), pen=(50, 250, 50, 50))
            except:
                print(self.parameter, 'cannot be evaluated')





import hashlib
import os
import time
import numpy as np
from PymoNNto.Exploration.StorageManager.StorageManager import *
from PymoNNto.Exploration.UI_Base import *
from PymoNNto.Exploration.Network_UI.Basic_Tabs.Helper.syntax import *


class Behavior_Parser_Object:
    md5_hash=''
    code=''
    name=''
    tags=None
    file=''
    base_class=''
    #time.ctime() for string conversion
    last_modification_time = None
    creation_time = None

    same_name=None
    same_tag=None
    base_or_derivative=None
    duplicates=None

    def __str__(self):
        return self.name+'('+self.md5_hash+')'

    def __repr__(self):
        return self.__str__()


def get_behaviors(file, base_classes=['Behavior']):

    result = []

    f = open(file, 'r')
    lines = f.readlines()

    class_found=False
    for line in lines:
        base_class_found=''
        if 'class' in line:
            for base in base_classes:
                if '('+base+')' in line:
                    base_class_found=base

        if base_class_found!='':
            class_found = True
            result.append([base_class_found])
        elif len(line)>0 and ord(line[0]) not in [10, 32]: #space or tab
                class_found = False

        if class_found:
            #if line.replace(chr(10),'').replace(chr(32),'')!=''): #do not add empty lines
            result[-1].append(line)#.replace('\n','')

    bpo_results=[]

    #if len(result)>0:
    #    print('')

    for r in result:

        bpo = Behavior_Parser_Object()

        bpo.file = file

        bpo.base_class = r.pop(0)

        name = r[0].replace('class ', '')
        name = name[0:min(name.find('('), name.find(':'))]
        bpo.name = name

        bpo.tags = []
        bpo.tags.append(name)

        for line in r:
            if 'self.add_tag(' in line and ')' in line:#self.add_tag('temporal_synapses')
                tag = line[line.index('('):line.index(')')]
                if len(tag)>=3:
                    bpo.tags.append(tag[2:-1])

        while r[-1].replace(chr(10),'').replace(chr(32),'') == '':#replace last empty lines
            r.pop(-1)
        bpo.code = ''.join(r)

        bpo.md5_hash = hashlib.md5(bpo.code.replace(chr(10),'').replace(chr(32),'').encode()).hexdigest()
        bpo.last_modification_time = os.path.getmtime(file)
        bpo.creation_time = os.path.getctime(file)

        bpo_results.append(bpo)

        #print(name)

    return bpo_results





class Behavior_UI(UI_Base):

    def __init__(self, path=get_root_folder()):#'.'
        super().__init__(title='Behavior overview')

        self.path=path

        self.hash_color_dict = {}

        file_list=[]
        for (dir_path, dir_names, file_names) in os.walk(path):
            if 'Data\\' not in dir_path:
                for f in file_names:
                    if f.endswith('.py'):
                        file_list.append(dir_path + '\\' + f)

        self.behaviors = []
        base_classes = [['Behavior']]
        while len(base_classes[-1]) > 0 and len(base_classes)<20:
            beh_temp = []
            for file_path in file_list:
                behaviors = get_behaviors(file_path, base_classes[-1])
                beh_temp+=behaviors
            base_classes.append([b.name for b in beh_temp])
            self.behaviors += beh_temp
            #print(beh_temp)

        self.sidebar.add_row()
        self.search_label = self.sidebar.add_widget(QLabel('Search:'), stretch=1)
        self.search = self.sidebar.add_widget(QLineEdit())
        self.search.setToolTip('filter filepaths, tags and behavior-names. Example: "MyBehaviorName -MyFolderName"')
        self.sidebar.set_parent_layout()

        def text_changed():
            self.update_file_list()

        self.search.textChanged.connect(text_changed)

        self.sidebar.add_widget(QLabel('Files with Behaviors:'), stretch=1)

        self.beh_list = self.sidebar.add_widget(QListWidget())

        self.sidebar.add_widget(QLabel('Versions:'),stretch=1)

        self.version_list = self.sidebar.add_widget(QListWidget(), stretch=50)



        self.behaviors=sorted(self.behaviors, key=lambda x: x.last_modification_time, reverse=True)

        print(len(self.behaviors))

        for b in self.behaviors:
            b.same_name=[b]
            b.same_tag=[]
            b.base_or_derivative=[]
            b.duplicates=[]
            for b2 in self.behaviors:
                if b is not b2:
                    if b.name==b2.name:
                        b.same_name.append(b2)
                    for t in b.tags:
                        if t in b2.tags and b2 not in b.same_tag:
                            b.same_tag.append(b2)
                    if b.base_class == b2.name or b.name == b2.base_class:
                        b.base_or_derivative.append(b2)
                    if b.md5_hash==b2.md5_hash:
                        b.duplicates.append(b2)

        self.update_file_list()

        self.add_code_tab()
        self.add_code_tab()
        self.add_code_tab()
        self.add_code_tab()

        def on_cb_click(event):
            self.version_list.clear()

            item=self.beh_list.currentItem()
            if item is not None:
                if hasattr(item, 'bpo'):
                    bpo = item.bpo
                    self.update_code_tab(bpo.code, bpo.file, bpo.last_modification_time)

                    self.version_list.addItem('same name:')

                    for b in bpo.same_name:
                        self.add_version_list_item(b)

                    self.version_list.addItem('')
                    self.version_list.addItem('same tag:')

                    for b in bpo.same_tag:
                        if b not in bpo.same_name:
                            self.add_version_list_item(b)

                    self.version_list.addItem('')
                    self.version_list.addItem('base or derivative:')

                    for b in bpo.base_or_derivative:
                        self.add_version_list_item(b)

                else:
                    file=item.text()
                    f = open(file, 'r')
                    code = f.read()
                    self.update_code_tab(code, file, os.path.getmtime(file))

        self.beh_list.currentItemChanged.connect(on_cb_click)

        def on_version_click(event):
            item=self.version_list.currentItem()
            if hasattr(item, 'bpo'):
                bpo=item.bpo
                self.update_code_tab(bpo.code, bpo.file, bpo.last_modification_time)

        self.version_list.currentItemChanged.connect(on_version_click)

    def add_code_tab(self):
        self.add_tab('Code', stretch=1)

        file_label = self.tab.add_widget(QLabel(''))
        self.tab.add_row()
        code_field = self.tab.add_widget(QPlainTextEdit())

        #self.btn = self.tab.add_widget(QPushButton('copy tab'), stretch=1)
        #def on_btn_click():
        #    self.add_code_tab()
        #self.btn.clicked.connect(on_btn_click)

        def on_txt_click(event):
            self.file_label=file_label
            self.code_field=code_field


        code_field.mouseReleaseEvent=on_txt_click

        code_field.highlight = PythonHighlighter(code_field.document(), ['neurons', 'synapses', 'network'])
        font = QFont()
        font.setPointSize(9)
        code_field.setFont(font)
        code_field.setStyleSheet("QPlainTextEdit { background-color: rgb(43, 43, 43); }")

        if not hasattr(self, 'file_label'):
            self.file_label=file_label
            self.code_field=code_field

    def searched(self, b):
        search_txt = self.search.text().split(' ')

        search_str_list = [stxt for stxt in search_txt if len(stxt)>0 and stxt[0]!='-']
        negative_search_str_list = [stxt[1:] for stxt in search_txt if len(stxt)>1 and stxt[0]=='-']

        search_found = True

        for negative_search_str in negative_search_str_list:
            if negative_search_str in b.name or negative_search_str in b.base_class or negative_search_str in b.file:
                search_found = False

        for search_str in search_str_list:
            if not (search_str in b.name or search_str in b.base_class or search_str in b.file):
                search_found = False

        return search_found

    def update_file_list(self):
        self.beh_list.clear()

        last_file = ''

        counter = 0

        for b in self.behaviors:

            if self.searched(b):

                counter += 1

                if last_file != b.file:
                    item=QListWidgetItem(b.file.replace(self.path,'.'))
                    item.setBackground(QColor(200,200,200))
                    self.beh_list.addItem(item)

                last_file = b.file

                count_info=''+str(len(b.same_name))+'|'+str(len(b.same_tag)-len(b.same_name)+1)+'|'+str(len(b.base_or_derivative))+''

                duplicates=''
                if len(b.duplicates)>0:
                    duplicates+='!'

                item=QListWidgetItem('          '+count_info+'    '+b.name+'    '+duplicates)
                item.bpo=b
                self.beh_list.addItem(item)

        self.search_label.setText('Search ('+str(counter)+'):')

    def update_code_tab(self, code, file, m_time):
        self.code_field.setPlainText(code)
        self.file_label.setText(file.replace(self.path,'.') + '    ' + time.ctime(m_time))

    def add_version_list_item(self, b):
        temp = '                   '
        if not self.searched(b):
            temp = '     [hidden]'

        new_item = QListWidgetItem(temp+'    '+b.name)  # + ' '+ b.md5_hash
        #if not self.searched(b):
        #    font=new_item.font()
        #    font.setStrikeOut(True)
        #    new_item.setFont(font)
        new_item.bpo = b
        if b.md5_hash not in self.hash_color_dict:
            self.hash_color_dict[b.md5_hash] = QColor(np.random.randint(50, 255),
                                                      np.random.randint(50, 255),
                                                      np.random.randint(50, 255))
        new_item.setBackground(self.hash_color_dict[b.md5_hash])
        self.version_list.addItem(new_item)


########################################################### Exception handling


def except_hook(cls, exception, traceback):
    sys.__excepthook__(cls, exception, traceback)

sys.excepthook = except_hook

if __name__ == '__main__':
    Behavior_UI().show()





def Behavior_UI():
    from PymoNNto.Exploration.Behavior_UI import Behavior_UI
    Behavior_UI().show()

def Execute_UI():
    from PymoNNto.Exploration.Execute_UI import Execute_UI
    Execute_UI().show()

def Overview_UI():
    from PymoNNto.Exploration.Overview_UI import Overview_UI
    Overview_UI().show()

def Evolution_UI():
    from PymoNNto.Exploration.Evolution.Evolution_UI import Evolution_UI
    Evolution_UI().show()



from PymoNNto.Exploration.Network_UI.Network_UI import *
from PymoNNto.Exploration.Network_UI.TabBase import *
import scipy.cluster.hierarchy as sch
import matplotlib.pyplot as plt

def generatePgColormap(cm_name):
    pltMap = plt.get_cmap(cm_name)
    colors = pltMap.colors
    colors = [c + [1.] for c in colors]
    positions = np.linspace(0, 1, len(colors))
    pgMap = pg.ColorMap(positions, np.array(colors)*255.0)
    return pgMap

class similarity_matrix_tab(TabBase):

    def __init__(self, title='similarity'):
        super().__init__(title)

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def update_labels(self):
        labels = self.left_axis_cb.get_selected_result()
        if labels is not None:
            if self.current_label_resort is not None:
                labels = list(np.flip(np.array(labels)[self.current_label_resort]))
            labels = [(0.5 + i, l) for i, l in enumerate(labels)]
            self.plot.showAxis('left')
            ax = self.plot.getAxis('left')
            ax.setTickFont(QFont("Courier"))
            ax.setWidth(100)
            ax.setTicks([labels])

        labels = self.right_axis_cb.get_selected_result()
        if labels is not None:
            if self.current_label_resort is not None:
                labels = list(np.flip(np.array(labels)[self.current_label_resort]))
            labels = [(0.5 + i, l) for i, l in enumerate(labels)]
            self.plot.showAxis('right')
            ax = self.plot.getAxis('right')
            ax.setTickFont(QFont("Courier"))
            ax.setWidth(100)
            ax.setTicks([labels])

    def update_plot(self):
        key = self.content_cb.get_selected_key()
        module = self.content_cb.get_selected_module()
        if module is not None:
            image, self.current_label_resort = module.get_cluster_matrix(key)
            if image is not None:
                self.image_item.setImage(np.fliplr(image))



    def on_selected_neuron_changed(self, Network_UI):
        self.left_axis_cb.change_main_object(Network_UI.selected_neuron_group())
        self.content_cb.change_main_object(Network_UI.selected_neuron_group())
        self.right_axis_cb.change_main_object(Network_UI.selected_neuron_group())
        self.update_plot()

    def initialize(self, Network_UI):
        self.similarity_matrix_tab = Network_UI.Next_Tab(self.title)

        self.image_item, self.plot = Network_UI.Add_Image_Item(True, False, title='', tooltip_message='afferent synapse weights of selected neuron')

        cmap = generatePgColormap('viridis')

        self.image_item.setLookupTable(cmap.getLookupTable())

        Network_UI.Next_H_Block()

        self.current_label_resort = None

        self.left_axis_cb = Network_UI.Add_element(Analytics_Results_Select_ComboBox(Network_UI.network.NeuronGroups[0], 'labeler', first_entry=''))
        self.content_cb = Network_UI.Add_element(Analytics_Results_Select_ComboBox(Network_UI.network.NeuronGroups[0], 'cluster_matrix_classifier'))
        self.right_axis_cb = Network_UI.Add_element(Analytics_Results_Select_ComboBox(Network_UI.network.NeuronGroups[0], 'labeler', first_entry=''))


        def update_btn_clicked(event):
            self.update_plot()
            self.update_labels()

        self.update_btn = Network_UI.Add_element(QPushButton('update'))
        self.update_btn.clicked.connect(update_btn_clicked)


    def update(self, Network_UI):
        if self.similarity_matrix_tab.isVisible():
            self.current_group = Network_UI.selected_neuron_group()






from PymoNNto import *

class CreateWeights(Behavior):

    def initialize(self, synapses):
        distribution = self.parameter('distribution', 'uniform(0.0,1.0)')  # ones
        density = self.parameter('density', 1.0)

        synapses.W = synapses.matrix(distribution, density=density) * synapses.enabled

        #  self.remove_autapses = self.parameter('remove_autapses', False) and synapses.src == synapses.dst

        if self.parameter('normalize', True):
            for i in range(10):
                synapses.W /= np.sum(synapses.W, axis=1)[:, None]
                synapses.W /= np.sum(synapses.W, axis=0)

            synapses.W *= self.parameter('nomr_fac', 1.0)


    def iteration(self, synapses):
        return
        #  if self.remove_autapses:
        #    np.fill_diagonal(synapses.W, 0.0)


from PymoNNto import *
from PymoNNto.Exploration.Evolution.Interface_Functions import *
import os
import sys

#def get_gene_string(gene):  # same implementation as interface_function get_gene_id
#    id = ''
#    for key, value in gene.items():
#        id += '#' + key + '$' + str(value)
#    return id + '#'

#def execute_local_file(slave_file, genome):
#    for arg in sys.argv:  # remove old
#        if 'genome=' in arg:
#            sys.argv.remove(arg)
#    sys.argv.append('genome=' + get_gene_id(genome))  # add new

#    py_file = open(slave_file, "r")
#    execution_string = py_file.read()
#    exec(execution_string)
#    py_file.close()

class Evolution_Device():#one device per thread

    def __init__(self, device_string, parent):
        self.device_string = device_string
        self.parent = parent
        self.current_gene = None

    def score_processing(self, evo_id):
        score = self.get_score(self.parent.name, evo_id)

        if score is not None:
            self.parent.new_score_event(evo_id, score)
        else:
            self.parent.error_event(evo_id, 'not able to load score (score=None)', False)

    def initialize_device_group(self):# override / called before initialize but only once per thread group
        return

    def initialize(self):# override
        return

    def main_loop_update(self):# override
        return

    def start(self):# override
        return

    def stop(self):# override
        return

    #def new_score_event(self, id, score):
    #    if genome.generation == self.parent.Breed_And_Select.generation:  # part of current execution? still relevant?
    #        self.parent.new_score_event(id, score)
    #    else:
    #        self.parent.error_event(id, 'processed gene result from previous generation', False)

    #def error_event(self, genome, message):
    #    if genome.generation == self.parent.Breed_And_Select.generation: #part of current execution? still relevant?
    #        self.parent.error_event(genome, message, True)
    #    else:
    #        self.parent.error_event(genome, message+' (from previous generation)')

    def get_score(self, evo_name, evo_id):
        sm = StorageManager(main_folder_name=evo_name, folder_name=get_gene_file(evo_name, evo_id), add_new_when_exists=False, use_evolution_path=False, last=True)#, print_msg=False
        return float(sm.load_param('score', default=None))



from PymoNNto.Exploration.Evolution.common_UI import *
from PymoNNto.Exploration.Evolution.PlotQTObjects import *

class Overview_UI(UI_Base):

    def update_order(self):
        for i in range(self.listwidget.count()):
            item = self.listwidget.item(i)
            #print(i, item.text())
            smg=self.interactive_scatter.get_smg(item.text())
            if smg is not None:
                smg.add_virtual_multi_parameter('index', i)

    def add_item(self, dir, folder):
        found = False
        for i in range(self.listwidget.count()):
            if self.listwidget.item(i).text() == dir:
                found = True

        if not found:
            item = QListWidgetItem(dir)
            item.folder = folder
            item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable)
            item.setCheckState(QtCore.Qt.Unchecked)
            self.listwidget.addItem(item)
            # dirs.append(dir)

    def refresh_dirs(self):
        #dirs = []
        folder = 'Plot_Project_Clones'
        for dir in os.listdir(get_epc_folder(folder)):
            if os.path.isdir(get_epc_folder(folder) + '/' + dir):
                self.add_item(dir, get_data_folder() + '/' + folder + '/' + dir + '/Data')

        folder = 'Evolution_Project_Clones'
        for dir in os.listdir(get_epc_folder(folder)):
            if os.path.isdir(get_epc_folder(folder) + '/' + dir):
                self.add_item(dir, get_data_folder() + '/' + folder + '/' + dir + '/Data')


        for dir in os.listdir(get_data_folder() + '/StorageManager/'):
            if os.path.isdir(get_data_folder() + '/StorageManager/' + dir):
                self.add_item(dir, get_data_folder())

        #self.listwidget.addItems(dirs)
        self.listwidget.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)

    def __init__(self):
        super().__init__(title='Plot Overview', create_sidebar=True)

        self.listwidget = QListWidget()

        self.add_tab(title='Plot')
        #self.Next_Tab('Plot')

        self.interactive_scatter = self.tab.add_widget(InteractiveScatter())

        self.refresh_dirs()

        def drop_event(ev):
            print(ev)
            super(QListWidget, self.listwidget).dropEvent(ev)
            self.update_order()
            self.interactive_scatter.refresh_data()

        self.update_order()

        self.listwidget.dropEvent = drop_event

        self.sidebar.add_widget(self.listwidget)

        def on_item_clicked(item):
            for i in range(self.listwidget.count()):
                item=self.listwidget.item(i)
                smg = StorageManagerGroup(item.text(), data_folder=item.folder)

                if item.checkState() == QtCore.Qt.Checked:
                    found_smg=self.interactive_scatter.get_smg(item.text())
                    if found_smg is None:
                        self.interactive_scatter.add_StorageManagerGroup(smg)
                        item.setBackground(QColor(smg.color[0],smg.color[1],smg.color[2]))
                else:
                    self.interactive_scatter.remove_StorageManagerGroup(smg)
                    item.setBackground(QColor(255,255,255))

            self.update_order()
            self.interactive_scatter.refresh_data()

        self.listwidget.itemClicked.connect(on_item_clicked)

        def on_item_doubleclicked(item):
            item = self.listwidget.itemFromIndex(item)
            smg = self.interactive_scatter.get_smg(item.text())

            if smg is not None:
                color = QColorDialog.getColor()
                print(color.red(), color.green(), color.blue())

                smg.color=(color.red(), color.green(), color.blue())
                item.setBackground(QColor(smg.color[0], smg.color[1], smg.color[2]))
                self.interactive_scatter.add_StorageManagerGroup(smg)
                self.interactive_scatter.refresh_data()

        self.listwidget.doubleClicked.connect(on_item_doubleclicked)

        btn=self.sidebar.add_widget(QPushButton('refresh'))

        def refresh_click():
            self.refresh_dirs()
            self.interactive_scatter.refresh_data()

        btn.clicked.connect(refresh_click)

        color_btn = self.sidebar.add_widget(QPushButton('color'))

        def color_click():
            self.interactive_scatter.axis_dialog('color')
            #param =
            #color_btn.setText('color: '+param)

        color_btn.clicked.connect(color_click)


########################################################### Exception handling


def except_hook(cls, exception, traceback):
    sys.__excepthook__(cls, exception, traceback)

sys.excepthook = except_hook


if __name__ == '__main__':
    Overview_UI().show()



from PymoNNto.Exploration.AnalysisModules.Weight_Classifier_Pre import *
from PymoNNto.Exploration.AnalysisModules.Weight_Classifier_Post import *
from PymoNNto.Exploration.AnalysisModules.Static_Classification import *


from PymoNNto import *

def behavior_test_environment(behavior, size=1, initialize=True, iteration=0):
    net = Network(tag='Network')

    if type(behavior) != dict:
        behavior={1:behavior}

    ng = NeuronGroup(tag='Neuron', net=net, behavior=behavior, size=size)

    if initialize:
        net.initialize()

    for i in range(iteration):
        net.simulate_iteration()

    return net, ng, behavior


#import matplotlib
#matplotlib.use('Qt5Agg')

from PyQt5 import QtCore
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import pyqtgraph as pg
import numpy as np
#from Exploration.Visualization.Visualization_Helper import *

from PyQt5.QtWidgets import QTabWidget
from PyQt5.QtCore import Qt, QPoint, QMimeData
from PyQt5.QtGui import QPixmap, QRegion, QDrag, QCursor

import sys
import PymoNNto

from functools import wraps
from pyqtgraph.graphicsItems.PlotItem import PlotItem

sys._excepthook = sys.excepthook

def exception_hook(exctype, value, traceback):
    print(exctype, value, traceback)
#    sys._excepthook(exctype, value, traceback)
    sys.exit(1)
sys.excepthook = exception_hook

def deprecated_warning(message):
    def inner_function(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            #raise Exception(message)
            print(message)
            return function(*args, **kwargs)
        return wrapper
    return inner_function


default_colors=[(0, 0, 0),(255, 0, 0),(0, 0, 255),(0, 250, 150),(255, 0, 255)]


class UI_Base(QApplication):

    def __init__(self, title="Network_Test", create_sidebar=True, create_tab_grid=True):#network
        super().__init__(sys.argv)

        path = PymoNNto.__file__.replace('__init__.py', '')+ 'icon3232.png'

        if 'win' in sys.platform and sys.platform != 'darwin':
            import ctypes
            myappid = 'mv.Pymonnto.ui.1'  # arbitrary string mycompany.myproduct.subproduct.version
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

        self._set_icon(path)

        self.reduced_layout = False

        self.main_window = QWidget()
        self.main_window.keyPressEvent = self.keyPressEvent
        self.main_window.keyReleaseEvent = self.keyReleaseEvent
        self.control_key_down=False

        self.init_QT_Window(title, create_sidebar, create_tab_grid)
        quit = QAction("Quit", self.main_window)
        quit.triggered.connect(self.main_window.closeEvent)
        self.closeeventtriggered = False

    def _set_icon(self, path):
        app_icon = QIcon()
        app_icon.addFile(path, QtCore.QSize(32, 32))
        self.setWindowIcon(app_icon)

    def show(self):
        self.main_window.show()
        sys.exit(self.exec_())

    def init_QT_Window(self, label, create_sidebar=True, create_tab_grid=True):
        self.width = 1200
        self.height = 600
        self.main_window.setWindowTitle(label)
        self.main_window.setGeometry(10, 40, self.width, self.height)

        self.main_h_layout = QHBoxLayout(self.main_window)

        self.main_split = QSplitter()
        self.main_h_layout.addWidget(self.main_split)

        if create_sidebar:
            self.sidebar = QRow_Column_Widget(self)
            self.main_split.addWidget(self.sidebar)

        self.add_tab_grid(create_tab_grid=create_tab_grid)


    def remove_tab(self, tab):
        for tab_widget in self._tab_widgets:
            indx = tab_widget.indexOf(tab)
            if indx!=-1:
                tab_widget.removeTab(indx)

    def add_tab_grid(self, create_tab_grid=True):

        self.tabs = MyTabWidget(new=True)

        self._tab_widgets = [self.tabs]

        if not create_tab_grid:
            self.main_split.addWidget(self.tabs)
        else:
            self.vsplit = QSplitter()
            self.vsplit.setOrientation(Qt.Vertical)
            #...
            self.hsplit1 = QSplitter()

            self.tabs2 = MyTabWidget(new=True)
            self._tab_widgets.append(self.tabs2)
            self.hsplit1.addWidget(self.tabs)
            self.hsplit1.addWidget(self.tabs2)
            self.hsplit1.setSizes([1, 0])

            self.hsplit2 = QSplitter()
            self.tabs3 = MyTabWidget(new=True)
            self._tab_widgets.append(self.tabs3)
            self.tabs4 = MyTabWidget(new=True)
            self._tab_widgets.append(self.tabs4)
            self.hsplit2.addWidget(self.tabs3)
            self.hsplit2.addWidget(self.tabs4)
            self.hsplit2.setSizes([1, 0])

            self.vsplit.addWidget(self.hsplit1)
            self.vsplit.addWidget(self.hsplit2)
            self.vsplit.setSizes([1, 0])

            self.main_split.addWidget(self.vsplit)
            self.main_split.setSizes([300, 800])

            self.tabs.currentChanged.connect(self._onTabChange)
            self.tabs2.currentChanged.connect(self._onTabChange)
            self.tabs3.currentChanged.connect(self._onTabChange)
            self.tabs4.currentChanged.connect(self._onTabChange)

    def _onTabChange(self, i):
        self.on_tab_change(i)

    def on_tab_change(self, i):
        return #overwrite

    def keyReleaseEvent(self, event):
        if event.key() == Qt.Key_Control:
            self.control_key_down=False

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Control:
            self.control_key_down=True

        if event.key() in [Qt.Key_N, Qt.Key_C, Qt.Key_T, Qt.Key_M]:
            self.text_input_dialog('Please enter Note:', 'set note', self.set_note, default_text='-')

        if event.key() == Qt.Key_W:
            indx = self.tabs.currentIndex()
            if indx >= 0:
                widget=self.tabs.currentWidget()
                self.tabs.removeTab(indx)
                widget.setParent(None)
                widget.setWindowTitle('Warning: experimental feature (closing/reattaching the window can lead to crashes)')
                widget.show()

        if event.key() == Qt.Key_Space:
            self.pause = not self.pause

        if event.key() == Qt.Key_1:
            self.vsplit.setSizes([100000, 0])
            self.hsplit1.setSizes([100000, 0])
            self.hsplit2.setSizes([100000, 0])

        if event.key() == Qt.Key_2:
            self.vsplit.setSizes([100000, 100000])
            self.hsplit1.setSizes([100000, 0])
            self.hsplit2.setSizes([100000, 0])

        if event.key() == Qt.Key_3:
            self.vsplit.setSizes([100000, 100000])
            self.hsplit1.setSizes([100000, 100000])
            self.hsplit2.setSizes([100000, 0])

        if event.key() == Qt.Key_4:
            self.vsplit.setSizes([100000, 100000])
            self.hsplit1.setSizes([100000, 100000])
            self.hsplit2.setSizes([100000, 100000])

        if event.key() == Qt.Key_5:
            self.vsplit.setSizes([100000, 100000])
            self.hsplit1.setSizes([100000, 0])
            self.hsplit2.setSizes([100000, 100000])

        if event.key() == Qt.Key_6:
            self.vsplit.setSizes([100000, 100000])
            self.hsplit1.setSizes([100000, 100000])
            self.hsplit2.setSizes([100000, 0])

        if event.key() == Qt.Key_7:
            self.vsplit.setSizes([100000, 0])
            self.hsplit1.setSizes([100000, 100000])
            self.hsplit2.setSizes([100000, 0])

    def set_note(self, note_str):
        self.main_window.setWindowTitle(self.main_window.windowTitle() + ' | ' + note_str)
        if note_str=='-':
            self.main_window.setWindowTitle(self.main_window.windowTitle().split('|')[0])

    def text_input_dialog(self, txt, btn_text, func, default_text=''):
        dlg = QDialog()
        dlg.setWindowTitle(txt)
        layout = QVBoxLayout()

        input_le = QLineEdit(default_text)
        layout.addWidget(input_le)

        def btn_clicked():
            func(input_le.text())
            dlg.close()

        btn = QPushButton(btn_text)
        btn.clicked.connect(btn_clicked)

        layout.addWidget(btn)
        dlg.setLayout(layout)
        dlg.resize(300, 50)
        dlg.exec()

    def show_message(self, title, message, btn_text='Ok'):
        dlg = QDialog()
        dlg.setWindowTitle(title)
        layout = QVBoxLayout()

        label = QLabel(message)
        layout.addWidget(label)

        def btn_clicked():
            dlg.close()

        btn = QPushButton(btn_text)
        btn.clicked.connect(btn_clicked)

        layout.addWidget(btn)
        dlg.setLayout(layout)
        dlg.resize(300, 50)
        dlg.exec()





    def add_tab(self, title='', stretch=100):
        self.tab = QRow_Column_Widget(self)

        def to_tab(event):
            self.tabs.addTab(self.tab, title)

        self.tab.closeEvent = to_tab
        self.tab.setWindowTitle(title)
        self.tabs.addTab(self.tab, title)
        #self.tab_current_layout = QVBoxLayout(main_tab_container)
        self.tab.add_row(stretch=stretch)
        return self.tab






    ################################################################################################
    #deprecated#####################################################################################
    ################################################################################################

    @deprecated_warning("Add_element function will be removed in coming versions. Please use tab/sidebar.add_widget instead.")
    def Add_element(self, elem, sidebar=False, stretch=1):
        return self.tab.add_widget(elem, sidebar=sidebar, stretch=stretch)
        #if sidebar:
        #    self.Add_Sidebar_Element(elem, stretch=stretch)#new stretch
        #else:
        #    self.current_H_block.addWidget(elem, stretch=stretch)
        #return elem

    @deprecated_warning("Add_plot function will be removed in coming versions. Please use tab/sidebar.add_plot instead.")
    def Add_plot(self, title=None, stretch=1, axisItems=None, x_label=None, y_label=None, always_show_axis_labels=False, tooltip_message=None):
        return self.tab.add_plot(title=title, stretch=stretch, axisItems=axisItems, x_label=x_label, y_label=y_label, tooltip_message=tooltip_message)
        #canvas = pg.GraphicsLayoutWidget()
        #canvas.setBackground((255, 255, 255))

        #if tooltip_message is not None:
        #    canvas.ci.setToolTip(tooltip_message)

        #self.Add_element(canvas, sidebar, stretch)

        #plt = canvas.addPlot(row=0, col=0, axisItems=axisItems)

        #if always_show_axis_labels or not self.reduced_layout:
        #    if x_label is not None:
        #        plt.getAxis('bottom').setLabel(text=x_label)
        #    if y_label is not None:
        #        plt.getAxis('left').setLabel(text=y_label)

        #if title is not None:
        #    plt.setLabels(title=title)
        #return plt

    @deprecated_warning("Add_Image_Item function will be removed in coming versions. Please use tab/sidebar.add_plot().add_image instead.")
    def Add_Image_Item(self, return_plot=False, sidebar=False, stretch=1, title=None, tooltip_message=None):
        plot = self.tab.add_plot(title=title, stretch=stretch, tooltip_message=tooltip_message)
        image_item = plot.add_image()

        #canvas = pg.GraphicsLayoutWidget()

        #if tooltip_message is not None:
        #    canvas.ci.setToolTip(tooltip_message)

        ##canvas.ci.layout.setContentsMargins(0, 0, 0, 0)
        ##canvas.ci.layout.setSpacing(0)

        #canvas.setBackground((255, 255, 255))

        #self.Add_element(canvas, sidebar, stretch)

        #plot = canvas.addPlot(row=0, col=0)
        ##plot.setContentsMargins(0,0,0,0)

        #if title is not None:
        #    plot.setLabels(title=title)

        #plot.hideAxis('left')
        #plot.hideAxis('bottom')
        #image_item = pg.ImageItem(np.random.rand(291, 291, 3))

        #plot.addItem(image_item)
        if return_plot:
            return image_item, plot
        else:
            return image_item

    @deprecated_warning("Add_plot_curve function will be removed in coming versions. Please use tab/sidebar.add_plot().add_curves instead.")
    def Add_plot_curve(self, title=None, return_plot=False, sidebar=False, stretch=1, number_of_curves=1, colors=[(0, 0, 0),(255, 0, 0),(0, 0, 255),(0, 250, 150),(255, 0, 255)], lines=[], names=[''], legend=False, x_label=None, y_label=None, return_list=False, always_show_axis_labels=False, tooltip_message=None):
        plot = self.tab.add_plot(title=title, stretch=stretch, x_label=x_label, y_label=y_label, tooltip_message=tooltip_message)
        curves = plot.add_curves(number_of_curves=number_of_curves, colors=colors, lines=lines, names=names, legend=legend)

        #plt = self.Add_plot(title, sidebar, stretch, tooltip_message=tooltip_message)

        #if always_show_axis_labels or not self.reduced_layout:
        #    if x_label is not None:
        #        plt.getAxis('bottom').setLabel(text=x_label)
        #    if y_label is not None:
        #        plt.getAxis('left').setLabel(text=y_label)

        #if legend:
        #    plt.addLegend()

        #for line in lines:
        #    plt.addLine(y=line,pen=(50,50,50,255))

        #curves=[]
        #for i in range(number_of_curves):
        #    curve = pg.PlotCurveItem([], name=names[i%len(names)], pen=colors[i%len(colors)])

        #    plt.addItem(curve)
        #    curves.append(curve)
        #if number_of_curves == 1 and not return_list:
        #    curves = curves[0]

        if return_plot:
            return curves, plot
        else:
            return curves

    @deprecated_warning("Next_Tab function will be removed in coming versions. Please use add_tab instead.")
    def Next_Tab(self, title='', stretch=100 , create_h_block=True):
        return self.add_tab(title=title, stretch=stretch)#, create_h_block=create_h_block
        #main_tab_container = QWidget()

        #def to_tab(event):
        #    self.tabs.addTab(main_tab_container, title)

        #main_tab_container.closeEvent = to_tab
        #main_tab_container.setWindowTitle(title)

        ##bl = QHBoxLayout(self.main_tab_container)
        ##bl.addWidget(self.net_info_text_label)
        #self.tabs.addTab(main_tab_container, title)
        #self.visualization_layout = QVBoxLayout(main_tab_container)
        #if create_h_block:
        #    self.Next_H_Block(stretch)
        #return main_tab_container

    @deprecated_warning("Next_H_Block function will be removed in coming versions. Please use tab/sidebar.add_row instead.")
    def Next_H_Block(self,stretch=100):
        return self.tab.add_row(stretch=stretch)
        #self.current_H_block = QHBoxLayout()
        #self.visualization_layout.addLayout(self.current_H_block, stretch=stretch)

    #@deprecated_warning("Add_New_Sidebar_Column function will be removed in coming versions.")
    #def Add_New_Sidebar_Column(self):
    #    self.sidebar_current_vertical_layout = QVBoxLayout()
    #    #self.sidebar_current_vertical_layout.setSpacing(0)
    #    #self.sidebar_current_vertical_layout.setContentsMargins(0, 0, 0, 0)
    #    self.sidebar_column_layout.addLayout(self.sidebar_current_vertical_layout, stretch=1)

    @deprecated_warning("Add_Sidebar_Spacing function will be removed in coming versions. Please use sidebar.add_widget(QLabel()) instead.")
    def Add_Sidebar_Spacing(self):
        self.add_sidebar_widget(QLabel())
        #self.Add_Sidebar_Element(QLabel())

    @deprecated_warning("Add_Sidebar_Element function will be removed in coming versions. Please use sidebar.add_widget(..., sidebar=True)")
    def Add_Sidebar_Element(self, elements=[], percentages=None, stretch=100, return_h_layout=False):
        #self.add_sidebar_widget(elements)

        if type(elements) is not list: elements=[elements]
        if percentages is None: percentages=[1 for _ in elements]

        #self.sidebar_hblock = QHBoxLayout()
        ##block.setSpacing(0)
        ##block.setMargin(0)
        ##block.setContentsMargins(0, 0, 0, 0)

        layout = self.sidebar.add_column(stretch=stretch)

        for element,p in zip(elements,percentages):
            self.sidebar.add_widget(element, stretch=p)
            #self.sidebar_hblock.addWidget(element,stretch=p)

        self.sidebar.set_parent_layout()

        #self.sidebar_current_vertical_layout.addLayout(self.sidebar_hblock, stretch=stretch)
        if return_h_layout:
            return layout
        else:
            return elements



    #def create_detail_label_pixmap(self):
    #    self.detail_qlabel = QLabel()
    #    self.detail_qlabel.setScaledContents(True)
    #    self.Add_Sidebar_Element(self.detail_qlabel,stretch=10)
    #    self.update_detail_image()

    #def update_detail_image(self):
    #    self.detail_pixmap = QPixmap(self.detail_qlabel.width(), self.detail_qlabel.height())
    #    painter = QPainter(self.detail_pixmap)
    #    painter.eraseRect(0, 0, self.detail_qlabel.width(), self.detail_qlabel.height())
    #    self.refresh_detail_image()

    #def refresh_detail_image(self):
    #    self.detail_qlabel.setPixmap(self.detail_pixmap)
    #    self.detail_qlabel.repaint()



    #def create_content_label_pixmap(self):
    #    self.content_qlabel = QLabel()
    #    self.content_qlabel.setScaledContents(True)
    #    self.visualization_layout.addWidget(self.content_qlabel, stretch=1)
    #    self.update_content_image()

    #def update_content_image(self):
    #    self.content_pixmap = QPixmap(self.content_qlabel.width(), self.content_qlabel.height())
    #    painter = QPainter(self.content_pixmap)
    #    painter.eraseRect(0, 0, self.content_qlabel.width(), self.content_qlabel.height())
    #    self.refresh_content_image()

    #def refresh_content_image(self):
    #    self.content_qlabel.setPixmap(self.content_pixmap)
    #    self.content_qlabel.repaint()


    #def draw_qimage(self, qimg, x, y, w, h, pixmap=None):
    #    if pixmap is None:
    #        pixmap=self.content_pixmap
    #    painter = QPainter(pixmap)
    #    painter.drawImage(QtCore.QRect(x, y, w, h), qimg, QtCore.QRect(0, 0, qimg.width(), qimg.height()))


    #def numpy_array_to_qimage(self, data, byte_count=3, alpha=255):
    #    data = data.astype(int)
    #    bytesPerLine = byte_count * data.shape[1]

    #    if byte_count == 1: format = QImage.Format_Grayscale8
    #    if byte_count == 3: format = QImage.Format_RGB888
    #    if byte_count == 4: format = QImage.Format_RGBA8888

    #    if byte_count == 4 and alpha < 255:
    #        data[:, :, 3] = alpha

    #    return QImage(np.require(data, np.uint8, 'C'), data.shape[1], data.shape[0], bytesPerLine, format)


class QRow_Column_Widget(QWidget):

    def __init__(self, _parent):
        super().__init__()
        self._parent = _parent
        self.current_layout = QVBoxLayout()
        self.setLayout(self.current_layout)


    def add_widget(self, widget, stretch=100, sidebar=False):
        if type(widget) == list:
            for w in widget:
                self.get_layout().addWidget(w, stretch=stretch)
        else:
            self.get_layout().addWidget(widget, stretch=stretch)
        return widget

    def add_graphics_layout_widget(self, tooltip_message=None, stretch=100):
        graphics_layout_widget = pg.GraphicsLayoutWidget()
        graphics_layout_widget.setBackground((255, 255, 255))

        if tooltip_message is not None:
            graphics_layout_widget.ci.setToolTip(tooltip_message)

        self.add_widget(graphics_layout_widget, stretch=stretch)
        return graphics_layout_widget


    def add_plot(self, title=None, tooltip_message=None, axisItems=None, x_label=None, y_label=None, stretch=100):
        graphics_layout_widget = self.add_graphics_layout_widget(tooltip_message=tooltip_message, stretch=stretch)

        #plot = graphics_layout_widget.addPlot(row=0, col=0, axisItems=axisItems)
        plot = PymoNNto_PlotItem(axisItems=axisItems)
        graphics_layout_widget.addItem(plot, row=0, col=0)

        plot._glayout=graphics_layout_widget

        if not self._parent.reduced_layout:

            if x_label is not None:
                plot.getAxis('bottom').setLabel(text=x_label)

            if y_label is not None:
                plot.getAxis('left').setLabel(text=y_label)

        if title is not None:
            plot.setLabels(title=title)

        return plot

    def add_row(self, stretch=100, add_to_parent=True):
        if add_to_parent:
            self.set_parent_layout()
        layout = QHBoxLayout()
        layout._parent = self.get_layout()
        layout._parent.addLayout(layout, stretch=stretch)
        self.set_layout(layout)
        return layout

    def add_column(self, stretch=100, add_to_parent=True):
        if add_to_parent:
            self.set_parent_layout()
        layout = QVBoxLayout()
        layout._parent = self.get_layout()
        layout._parent.addLayout(layout, stretch=stretch)
        self.set_layout(layout)
        return layout

    def set_parent_layout(self, root=False):
        current = self.get_layout()
        if hasattr(current, '_parent') and current._parent is not None:
            self.set_layout(self.get_layout()._parent)
            if root:
                self.set_parent_layout(root=root)



    def get_layout(self):
            return self.current_layout

    def set_layout(self, layout):
        self.current_layout = layout




class PymoNNto_PlotItem(PlotItem):

    def add_curves(self, number_of_curves=1, colors=default_colors, lines=[], names=[''], legend=False, return_list=False):

        for line in lines:
            self.addLine(y=line, pen=(50, 50, 50, 255))

        curves = []
        for i in range(number_of_curves):
            curve = pg.PlotCurveItem([], name=names[i % len(names)], pen=colors[i % len(colors)])
            self.addItem(curve)
            curves.append(curve)

        #if number_of_curves == 1 and not return_list:
        #    curves = curves[0]

        if legend:
            self.addLegend()

        return curves

    def add_image(self, no_border=True):
        if no_border:
            self._glayout.ci.layout.setContentsMargins(0, 0, 0, 0)
            self._glayout.ci.layout.setSpacing(0)

        self.hideAxis('left')
        self.hideAxis('bottom')
        image_item = pg.ImageItem()#np.zeros((100, 100, 3))
        self.addItem(image_item)
        return image_item

    def add_text(self, txt):
        text_item = pg.TextItem(txt)
        self.addItem(text_item)
        return text_item






########################

from PyQt5 import QtCore, QtGui, QtWidgets


class TabBar(QtWidgets.QTabBar):
    def tabSizeHint(self, index):
        s = QtWidgets.QTabBar.tabSizeHint(self, index)
        s.transpose()
        return s

    def paintEvent(self, event):
        painter = QtWidgets.QStylePainter(self)
        opt = QtWidgets.QStyleOptionTab()

        for i in range(self.count()):
            self.initStyleOption(opt, i)
            painter.drawControl(QtWidgets.QStyle.CE_TabBarTabShape, opt)
            painter.save()

            s = opt.rect.size()
            s.transpose()
            r = QtCore.QRect(QtCore.QPoint(), s)
            r.moveCenter(opt.rect.center())
            opt.rect = r

            c = self.tabRect(i).center()
            painter.translate(c)
            painter.rotate(90)
            painter.translate(-c)
            painter.drawControl(QtWidgets.QStyle.CE_TabBarTabLabel, opt);
            painter.restore()

class ProxyStyle(QtWidgets.QProxyStyle):
    def drawControl(self, element, opt, painter, widget):
        if element == QtWidgets.QStyle.CE_TabBarTabLabel:
            ic = self.pixelMetric(QtWidgets.QStyle.PM_TabBarIconSize)
            r = QtCore.QRect(opt.rect)
            w =  0 if opt.icon.isNull() else opt.rect.width() + self.pixelMetric(QtWidgets.QStyle.PM_TabBarIconSize)
            r.setHeight(opt.fontMetrics.width(opt.text) + w)
            r.moveBottom(opt.rect.bottom())
            opt.rect = r
        QtWidgets.QProxyStyle.drawControl(self, element, opt, painter, widget)

################





class MyTabWidget(QTabWidget):

   def __init__(self, parent=None, new=None):
      super().__init__(parent)

      if 'sideTab' in sys.argv:
          self.setTabBar(TabBar(self))
          self.setTabPosition(QtWidgets.QTabWidget.West)

      self.setAcceptDrops(True)
      self.tabBar().setMouseTracking(True)
      self.setMovable(True)
      if new:
         MyTabWidget.setup(self)

   def __setstate__(self, data):
      self.__init__(new=False)
      self.setParent(data['parent'])
      for widget, tabname in data['tabs']:
         self.addTab(widget, tabname)
      MyTabWidget.setup(self)

   def __getstate__(self):
      data = {
         'parent' : self.parent(),
         'tabs' : [],
      }
      tab_list = data['tabs']
      for k in range(self.count()):
         tab_name = self.tabText(k)
         widget = self.widget(k)
         tab_list.append((widget, tab_name))
      return data

   def setup(self):
       pass

   def mouseMoveEvent(self, e):
       if e.buttons() != Qt.RightButton:
           return

       globalPos = self.mapToGlobal(e.pos())
       tabBar = self.tabBar()
       posInTab = tabBar.mapFromGlobal(globalPos)
       index = tabBar.tabAt(e.pos())
       tabBar.dragged_content = self.widget(index)
       tabBar.dragged_tabname = self.tabText(index)
       tabRect = tabBar.tabRect(index)

       pixmap = QPixmap(tabRect.size())
       tabBar.render(pixmap, QPoint(), QRegion(tabRect))
       mimeData = QMimeData()

       drag = QDrag(tabBar)
       drag.setMimeData(mimeData)
       drag.setPixmap(pixmap)

       cursor = QCursor(Qt.OpenHandCursor)

       drag.setHotSpot(e.pos() - posInTab)
       drag.setDragCursor(cursor.pixmap(), Qt.MoveAction)
       drag.exec_(Qt.MoveAction)

   def dragEnterEvent(self, e):
       e.accept()
       # self.parent().dragged_index = self.indexOf(self.widget(self.dragged_index))

   def dragLeaveEvent(self, e):
       e.accept()

   def dropEvent(self, e):
       if e.source().parentWidget() == self:
           return

       e.setDropAction(Qt.MoveAction)
       e.accept()
       tabBar = e.source()
       self.addTab(tabBar.dragged_content, tabBar.dragged_tabname)

#tqt = TREN_QT_UI_Base(None)
#tqt.show()

#########################
#########################Manual
#########################

'''
self.create_detail_label_pixmap()
self.create_content_label_pixmap()

for i in range(5):
    self.Add_Sidebar_Element([QPushButton('refresh', self.main_window), QPushButton('refresh', self.main_window)])


def onclick(event):
    self.update_detail_image()
    self.update_content_image()

    painter = QPainter(self.detail_pixmap)
    pen = QPen(QColor(255, 255, 0), 3)
    painter.setPen(pen)
    painter.drawLine(10, 10, 100, 100)
    painter.drawRect(painter.clipBoundingRect())

    self.refresh_content_image()
    self.refresh_detail_image()

    # print('adssgdf')


for i in range(5):
    self.Add_Sidebar_Element(QPushButton('refresh', self.main_window))[0].clicked.connect(onclick)

self.Add_Sidebar_Spacing()

self.Add_New_Sidebar_Column()

self.Add_Sidebar_Spacing()

for i in range(5):
    self.Add_Sidebar_Element(QPushButton('refresh', self.main_window))
'''

